# -*- coding: utf-8 -*-
"""
Created on Mon Aug 22 15:24:49 2016

@author: akibdpt
"""
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import

from builtins import super
"""
from future import standard_library
standard_library.install_aliases()
"""
from .multiplyForm import multiplyForm

"""
 * INTERVAL is a behavioral category for Sorts (see 'Sort.java').
 * An interval sort is a multiply-associated sort for which
 * ...
 * Therefore, the IntervalForm class extends on the MultiplyForm
 * class (see 'MultiplyForm.java') and implements the algebra methods
 * from the Form interface (see 'Form.java').
 *
 * ASSUMPTIONS: the following algebraic property is assumed on forms
 *   a + b = b <=> a < b
 * 
"""
class interval(multiplyForm):
    type = 'form-interval'
    
    def __init__(self, sort, *args, **kwargs):
        super(interval, self).__init__(sort)
    
    """
    print out the form for easier assessment
    structure:
    |sort name: n elements
     _ individual 1
     _ individual 2
         ...
     _ individual n
      |attribute1 sort name: m elements
       _ attribute1 1
      |attribute2 sort name: p elements
       _ attribute1 2
    """
    def printE(self, string = '', *args, **kwargs):
        print(string + '|' + self.sort.name+':',len(self.elements),'elements')
        for ind in self.elements:
            print(string, end = ' ')
            print(ind.printVal())
            if ind.hasAttribute():
                ind.attribute.printE(string+'  ')
            """
            print(string, ind.printVal())
            """
        return
        
    """
    inserting an element into the form
    """
    
    def insert(self, ind):
        if not self.sort == ind.sort:
            raise ValueError('Inserting an object of sort '+ind.sort.name+' into a form of sort '+self.sort.name)
        if ind.type == self.type:
            for i in ind:
                self.insert(i)
            return
        thisInd = self.first()
        index = 0
        while thisInd is not None and thisInd.compareInterval(ind) < 1:
            if thisInd.compareInterval(ind) == 0:
                """
        while thisInd is not None:
            compareI = thisInd.compareInterval(ind)
            if compareI < 1:
                break
            if compareI == 0:
        """
                if thisInd.hasAttribute() or ind.hasAttribute():
                    thisInd.attribute.sum(ind.attribute)
                return
            else:
                index = index+1
                try:
                    thisInd = self.elements[index]
                except:
                    thisInd = None
        self.add(index, ind)
        self.maximal = False
        return
                
    
    def oneStepForward(self, index):
        current = self.get(index)
        nextInd = self.get(index+1)
        if nextInd is not None and current.compareValue(nextInd) == 0 and current.touches(nextInd) and nextInd.attribute == current.attribute:
            self.elements[index+1] = nextInd.combine(current)
            if current.hasAttribute():
                self.elements[index+1].addAttribute(current.attribute)
                self.elements[index+1].attribute.sum(nextInd.attribute)
            nextInd = self.elements[index+1]
            self.remove(current)
        else:
            index += 1
        return index
        
    """
    maximalize the form:
    _remove Nil elements
    _reorder the elements in the form in ascending order
    """
    
    def maximalize(self, ignoreMaximal = False):
        if self.maximal and ignoreMaximal == False:
            return
        if len(self.elements) == 0:
            return
        ind = self.first()
        index = 1
        nextInd = self.get(index)
        
        disorganized = []
        while nextInd is not None:
            if nextInd.isNil():
                self.remove(nextInd)
                nextInd = self.get(index)
                continue
            elif ind.isNil():
                self.remove(ind)
                ind = nextInd
                nextInd = self.get(index)
                continue
            cmp = ind.compareValue(nextInd)
            if cmp == 0:
                cmpI = ind.compareInterval(nextInd)
                if cmpI == 0:
                    if ind.hasAttribute() or nextInd.hasAttribute():
                        ind.attribute.sum(nextInd.attribute)                
                    self.remove(nextInd)
                    nextInd = self.get(index)
                elif ind.touches(nextInd) :
                    if (not ind.hasAttribute() and not nextInd.hasAttribute()) or ind.attribute == nextInd.attribute:
                        self.remove(nextInd)
                        self.elements[index-1] = ind.combine(nextInd)
                        if ind.hasAttribute():
                            self.elements[index-1].addAttribute(ind.attribute)
                        ind = self.elements[index-1]
                        nextInd = self.get(index)
                    else:
                        if cmpI == 1:
                            self.remove(nextInd)
                            disorganized.append(nextInd)
                            nextInd = self.get(index)
                        else:
                            ind = nextInd
                            index += 1
                            nextInd = self.get(index)
                elif ind.disjoint(nextInd):
                    if cmpI == 1:
                        self.remove(nextInd)
                        disorganized.append(nextInd)
                        nextInd = self.get(index)
                    else:
                        ind = nextInd
                        index += 1
                        nextInd = self.get(index)
                    
                elif ind.contains(nextInd):
                    if (not ind.hasAttribute() and not nextInd.hasAttribute()) or ind.attribute == nextInd.attribute:
                        self.remove(nextInd)
                        nextInd = self.get(index)
                    else:
                        temp = ind.complement(nextInd)
                        if ind.hasAttribute() or nextInd.hasAttribute():
                            nextInd.attribute.sum(ind.attribute)
                        toBeRemoved = ind             
                        
                        if temp[0].compareInterval(nextInd) == -1:
                            ind = temp[0]
                        else: 
                            ind = nextInd
                            
                        self.remove(toBeRemoved)
                        
                        for i in temp:
                            self.insert(i)
                        nextInd = self.get(index)
                        
                elif nextInd.contains(ind):
                    if (not ind.hasAttribute() and not nextInd.hasAttribute()) or ind.attribute == nextInd.attribute:
                        self.remove(ind)
                        ind = nextInd
                        nextInd = self.get(index)
                    else:
                        temp = nextInd.complement(ind)
                        if ind.hasAttribute() or nextInd.hasAttribute():
                            ind.attribute.sum(nextInd.attribute)
                        if temp[0].compareInterval(ind) == -1:
                            ind = temp[0]
                        self.remove(nextInd)
                        for i in temp:
                            self.insert(i)
                        nextInd = self.get(index)
                        
                else:
                    if (not ind.hasAttribute() and not nextInd.hasAttribute()) or ind.attribute == nextInd.attribute:
                        self.remove(nextInd)
                        self.elements[index-1] = ind.combine(nextInd)
                        if ind.hasAttribute():
                            self.elements[index-1].addAttribute(ind.attribute)
                        ind = self.elements[index-1]
                        nextInd = self.get(index)
                    else:
                        temp1 = ind.complement(nextInd)[0]
                        temp2 = nextInd.complement(ind)[0]
                        ind = self.elements[index-1] = ind.common(nextInd)
                        if ind.hasAttribute() or nextInd.hasAttribute():
                            ind.attribute.sum(nextInd.attribute)
                        self.remove(nextInd)
                        self.insert(temp1)
                        self.insert(temp2)
                        nextInd = ind
                        ind = temp1
            elif cmp == 1:                
                self.remove(nextInd)
                disorganized.append(nextInd)
                nextInd = self.get(index)
            else:
                ind = nextInd
                index += 1
                nextInd = self.get(index)
        for ins in disorganized:
            self.insert(ins)
            
        for ind in self.elements:
            if ind.hasAttribute():
                ind.attribute.maximalize()    
                
        index = 0
        while index < len(self.elements):
            index = self.oneStepForward(index)
            
        self.maximal = True
        return
    
    
    """
    sum two forms (OR)
    e.g. formA: a, b, c
         formB:    b, c, d
        _formA.sum(formB)
      -->formA: a, b, c, d
    """
    
    def sum(self, other):
        if self is other:
            raise ValueError('Operation on two instances of the same form')
        if not self.sort.__eq__(other.sort):
            raise ValueError('Operation on two forms of different sorts '+str(self.sort)+'  and '+str(other.sort))
        
        for i in other.elements:
            self.insert(i)
                
        self.maximalize()
        other.maximalize()
        return

    """
    difference of two forms
    e.g. formA: a, b, c
         formB:    b, c, d
        _formA.difference(formB)
      -->formA: a
    """
    
    def difference(self, other):
        if self is other:
            raise ValueError('Operation on two instances of the same form')
        if not self.sort.__eq__(other.sort):
            raise ValueError('Operation on two forms of different sorts '+str(self.sort)+'  and '+str(other.sort))
        
        self.maximalize()
        other.maximalize()
            
        thisInd = self.first()
        otherInd = other.first()
        thisIndex = 0
        otherIndex = 0

        while thisInd is not None and otherInd is not None:
            cmp = thisInd.compareValue(otherInd) 
            if cmp == 0:
                cmpI = thisInd.compareInterval(otherInd)
                if thisInd.disjoint(otherInd) or thisInd.touches(otherInd):
                    if  cmpI == 1:                
                        otherIndex += 1
                        otherInd = other.get(otherIndex)
                    else:     
                        thisIndex += 1
                        thisInd = self.get(thisIndex)
                        
                elif cmpI == 0:
                    if thisInd.hasAttribute() or otherInd.hasAttribute():
                        thisInd.attribute.difference(otherInd.attribute)
                    if not thisInd.hasAttribute():
                        self.remove(thisInd)
                        thisInd = self.get(thisIndex)
                    otherIndex += 1
                    otherInd = other.get(otherIndex)
                    
                elif thisInd.contains(otherInd):
                    temp = thisInd.complement(otherInd)
                    if thisInd.hasAttribute() or otherInd.hasAttribute():
                        for i in temp:
                            i.addAttribute(thisInd.attribute)
                        thisInd.attribute.difference(otherInd.attribute)
                        if thisInd.hasAttribute():
                            self.elements[thisIndex] = thisInd.common(otherInd)
                            self.elements[thisIndex].addAttribute(thisInd.attribute)
                            thisInd = self.elements[thisIndex]
                            for i in temp:
                                self.insert(i)
                            thisIndex += 1
                            if thisInd == self.get(thisIndex):
                                thisIndex += 1
                            thisInd = self.get(thisIndex)
                        else:
                            self.remove(thisInd)
                            for i in temp:
                                self.insert(i)
                            thisInd = self.get(thisIndex)
                    elif not thisInd.hasAttribute():
                        self.remove(thisInd)
                        for i in temp:
                            self.insert(i)
                        thisInd = self.get(thisIndex)
   
                elif otherInd.contains(thisInd):
                    if thisInd.hasAttribute() or otherInd.hasAttribute():
                        thisInd.attribute.difference(otherInd.attribute)
                    if not thisInd.hasAttribute():
                        self.remove(thisInd)
                        thisInd = self.get(thisIndex)
                    else:
                        thisIndex += 1
                        thisInd = self.get(thisIndex)
                elif cmpI == -1:   
                    temp = thisInd.complement(otherInd)
                    if thisInd.hasAttribute() or otherInd.hasAttribute():
                        for i in temp:
                            i.addAttribute(thisInd.attribute)
                        thisInd.attribute.difference(otherInd.attribute)
                    if thisInd.hasAttribute():
                        self.elements[thisIndex] = thisInd.common(otherInd)
                        self.elements[thisIndex].addAttribute(thisInd.attribute)
                        thisInd = self.elements[thisIndex]
                        self.insert(temp[0])
                        thisIndex += 2
                        thisInd = self.get(thisIndex)
                    else:
                        self.remove(thisInd)
                        self.insert(temp[0])
                        thisIndex += 1
                        thisInd = self.get(thisIndex)
                else:
                    temp = thisInd.complement(otherInd)
                    if thisInd.hasAttribute() or otherInd.hasAttribute():
                        for i in temp:
                            i.addAttribute(thisInd.attribute)
                        thisInd.attribute.difference(otherInd.attribute)
                    if thisInd.hasAttribute():
                        self.elements[thisIndex] = thisInd.common(otherInd)
                        self.elements[thisIndex].addAttribute(thisInd.attribute)
                        thisInd = self.elements[thisIndex]
                        thisIndex += 1
                    else:
                        self.remove(thisInd)
                    self.insert(temp[0])
                    thisInd = self.get(thisIndex)
            elif cmp == 1: 
                otherIndex += 1
                otherInd = other.get(otherIndex)
            else:                      
                thisIndex += 1
                thisInd = self.get(thisIndex)

        index = 0
        
        while index < len(self.elements):
            index = self.oneStepForward(index)
        return
        
    """
    product of two forms (AND)
    e.g. formA: a, b, c
         formB:    b, c, d
        _formA.product(formB)
      -->formA:    b, c
    """
    def product(self, other):
        if self is other:
            raise ValueError('Operation on two instances of the same form')
        if not self.sort.__eq__(other.sort):
            raise ValueError('Operation on two forms of different sorts '+str(self.sort)+'  and '+str(other.sort))
            
        self.maximalize()
        other.maximalize()
                       
        thisInd = self.first()
        nextInd = self.next(thisInd)
        otherInd = other.first()
        while thisInd is not None and otherInd is not None:
            if thisInd.compareValue(otherInd) == 0:
                if thisInd.disjoint(otherInd) or thisInd.touches(otherInd):
                    if thisInd.compareInterval(otherInd) == 1:                
                        otherInd = other.next(otherInd)
                    else:
                        self.remove(thisInd)
                        thisInd = nextInd
                        nextInd = self.next(thisInd)
                else:
                    temp = thisInd.duplicateInd()
                    temp = temp.common(otherInd)
                    if thisInd.hasAttribute() and otherInd.hasAttribute():
                        temp.addAttribute(thisInd.attribute)
                        temp.attribute.product(otherInd.attribute)
                        if temp.hasAttribute():
                            self.insert(temp)
                    else:
                        self.insert(temp)
                        
                    if thisInd.compareEnd(otherInd) < 1:
                        self.remove(thisInd)
                        thisInd = nextInd
                        nextInd = self.next(thisInd)
                    else:
                        otherInd = other.next(otherInd)
  
            elif thisInd.compareValue(otherInd) == 1: 
                otherInd = other.next(otherInd)
            else:         
                toBeRemovedInd = thisInd
                thisInd = self.next(thisInd)
                self.remove(toBeRemovedInd)
        while thisInd is not None:   
            self.remove(thisInd)
            thisInd = nextInd
            nextInd = self.next(thisInd)
        return

        
                
    """
    check if this form contains a certain individual
    """
    def containsInd(self, ind):
        com = None
        if self.sort == ind.sort:
            for i in self.elements:
                if not i.compareValue(ind) == 0:
                    continue
                if i.containsInd(ind):
                    return True
                elif i.containsPoint(ind.start) and (not i.end == ind.start) and self.sort.type == 'sort-attribute':
                    if ind.hasAttribute():
                        if not ind.attribute.checkPartOf(i.attribute):                    
                            continue
                    com = i
                elif com is not None:
                    if i.start == com.end and i.containsPoint(ind.end) and ind.attribute.checkPartOf(i.attribute):
                        com = com.combine(i)
                        if com.contains(ind):
                            return True
                        return False
                    elif i.start == com.end and ind.contains(i) and ind.attribute.checkPartOf(i.attribute):
                        com = com.combine(i)
        return False
        
        
    """
    check if this form is a part of other form
    """
    def partOf(self, other):
        if self is other:
            raise ValueError('Operation on two instances of the same form')
        if not self.sort == other.sort:
            if other.sort.contains(self.sort):
                for i in other:
                    if i.sort == self.sort or i.sort.contains(self.sort) and self.partOf(i):
                        return True
                return False
            raise ValueError('Operation on two forms of different sorts '+str(self.sort)+'  and '+str(other.sort))
        self.maximalize()
        other.maximalize()
        thisInd = self.first()
        otherInd = other.first()
        for thisInd in self.elements:
            check = False
            com = None
            for otherInd in other.elements:
                if not otherInd.compareValue(thisInd) == 0:
                    continue
                if thisInd == otherInd or otherInd.contains(thisInd):
                    if thisInd.hasAttribute() and not thisInd.attribute.partOf(otherInd.attribute):
                        return False
                    check = True
                    break

                elif otherInd.containsPoint(thisInd.start):
                    if thisInd.hasAttribute():
                        if not thisInd.attribute.partOf(otherInd.attribute):
                            continue
                    com = otherInd.duplicateInd()
                elif com is not None:
                    if (otherInd.start == com.end and otherInd.containsPoint(thisInd.end) 
                    and self.sort.type == 'sort-attribute' and thisInd.attribute.partOf(otherInd.attribute)):
                        com = com.combine(otherInd)
                        if com.contains(thisInd):
                            check = True
                            break
                    elif (otherInd.start == com.end and thisInd.contains(otherInd) 
                    and self.sort.type == 'sort-attribute' and thisInd.attribute.partOf(otherInd.attribute)):
                        com = com.combine(otherInd)

            if check is False:
                return False
        return True
