# -*- coding: utf-8 -*-
"""
Created on Mon Jan  8 09:43:00 2018

@author: akibdpt
"""
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import range
from builtins import super
from builtins import str

from .individual import individual
from .point import point2D, point3D
from .plane import plane3D
from ..struct.coordinate import c0, c1, c0p5, coordinate
from ..struct.vector import vector2D, vector3D
from ..transform.affineTransform import affineTransform2D, affineTransform3D
from ..transform.graphTransform import graphTransform
from ..io.parseReader import ParsingError

v0 = vector2D(0,0)

class bezierCurve2D(individual):
    type = 'ind-bezierCurve2D'
    def __init__(self, sort, p1, p2, p3, isFunction = False, *args, **kwargs):
        super(bezierCurve2D, self).__init__(sort)
        if isFunction:
            self.a = p1
            self.b = p2
            self.c = p3
            self.start = p3
            self.controlPoint = self.b.scale(c0p5).add(p3)
            self.end = self.a.subtract(p3).add(self.controlPoint).add(self.controlPoint)
        else:
            self.a = p3.add(p1).subtract(p2.scale(2))
            self.b = (p2.subtract(p1)).scale(2)
            self.c = p1
            self.start = p1
            self.controlPoint = p2
            self.end = p3
        self.A = 4 * (self.a.x ** 2 + self.a.y ** 2)
        self.B = 4 * (self.a.x * self.b.x + self.a.y * self.b.y)
        self.C = self.b.x ** 2 + self.b.y ** 2
        self.D = 2 * (self.a.y * self.b.x - self.a.x * self.b.y)
        self.evaluateCurve()

        
    def vertex(self):
        t = -self.B/(2*self.A)
        return self.pos(t)
        
    def evaluateCurve(self):
        t = -self.B/(2*self.A)
        vg = self.vectGradient(t)
        if vg < v0 or (vg == v0 and self.vectGradient(t-1) < v0):
            self.start, self.end = self.end, self.start
            self.b = self.a.scale(-2).subtract(self.b)
            self.c = self.start
            self.B = 4 * (self.a.x * self.b.x + self.a.y * self.b.y)
            self.C = self.b.x ** 2 + self.b.y ** 2
            self.D = 2 * (self.a.y * self.b.x - self.a.x * self.b.y)
        
    def __str__(self):
        return '<'+str(self.start)+','+str(self.controlPoint)+','+str(self.end)+'>'
        return str(self.a)+'*t^2 + '+str(self.b)+'*t + '+str(self.c)
        
    def getValue(self):
        return str(self)

    def printVal(self):
        return '<'+str(self.start)+','+str(self.controlPoint)+','+str(self.end)+'>'

    def printEquation(self):
        return str(self.a)+'*t^2 + '+str(self.b)+'*t + '+str(self.c)

    def printRhino(self, displacement = 0):
        s = ''
        s += '_Curve '+str(self.start.x + displacement)+','+str(self.start.y)+'  '+str(self.controlPoint.x + displacement)+','+str(self.controlPoint.y)+'  '+str(self.end.x + displacement)+','+str(self.end.y)+' EnterEnd'
        return s
        
    def equals(self, other):
        return self.a == other.a and self.b == other.b and self.c == other.c
        
    def isNil(self):
        return self.start.x == self.controlPoint.x == self.end.x and self.start.y == self.controlPoint.y == self.end.y
        
    def compareValue(self, other, params = None):
        
        if params is None:
            l = 1/(((other.A**(3/2)) * (self.D**2))/((self.A**(3/2)) * (other.D**2)))
            phi = (self.A * other.D)/(other.A * self.D)
            v = (self.B * other.D - other.B * self.D)/(other.A * self.D * 2)
        else:
            l, phi, v = params
        c = self.pos(0).compareValue(other.pos(v).scale(l))
        if c != 0:
            return c
        return self.pos(1).compareValue(other.pos(phi+v).scale(l))
        
    def compareInterval(self, other):
        l = 1/(((other.A**(3/2)) * (self.D**2))/((self.A**(3/2)) * (other.D**2)))
        phi = (self.A * other.D)/(other.A * self.D)
        v = (self.B * other.D - other.B * self.D)/(other.A * self.D * 2)
        c = self.compareValue(other, params = (l,phi,v))
        if c != 0:
            return c
        c = v.compareValue(c0)
        if c != 0:
            return c
        c = (phi+v).compareValue(c1)
        if c != 0:
            return c
        return 0
        
    def pos(self, t):
        tc = coordinate(t)
        return self.a.scale(tc**2).add(self.b.scale(tc)).add(self.c)
        
    def curvature(self, t):
        tc = coordinate(t)
        return self.D / ((self.A * (tc ** 2)+ self.B * tc + self.C)**(3/2))
                
    def gradient(self, t):
        dx = 2 * self.a.x * t + self.b.x
        dy = 2 * self.a.y * t + self.b.y
        if dx == 0:
            return None
        return dy/dx
        
    def vectGradient(self, t):
        dx = 2 * self.a.x * t + self.b.x
        dy = 2 * self.a.y * t + self.b.y
        return vector2D(dx, dy)
                
    def codescriptor(self):
        point = point2D(self.sort, self.vertex())
        point.assoc = self
        return point
        
    def solveT(self, point):
        a = self.a.x
        b = self.b.x
        c = self.c.x - point.x
        
        if a == 0:
            x1 = -c/b
            x2 = None
        else:
            d = b*b-4*a*c # discriminant
            if d < 0:
                return None
            elif d == 0:
                x1 = (-b)/(2*a)
                x2 = None
            else:
                x1 = (-b+d.sqrt())/(2*a)
                x2 = (-b-d.sqrt())/(2*a)
            
        a = self.a.y
        b = self.b.y
        c = self.c.y - point.y
        if a == 0:
            y1 = -c/b
            y2 = None
        else:
            d = b**2-4*a*c
            
            if d < 0:
                return None
            elif d == 0:
                y1 = (-b)/(2*a)
                y2 = None
            else:
                y1 = (-b+d.sqrt())/(2*a)
                y2 = (-b-d.sqrt())/(2*a)
        x1r = round(x1.value,3)
        y1r = round(y1.value,3)
        if x1r == y1r:
            return x1
        if y2 is not None and x1r == round(y2.value,3):
            return x1
        if x2 is not None:
            x2r = round(x2.value,3)
            if x2r == y1r:
                return x2
            if y2 is not None and x2r == round(y2.value,3):
                return x2
        return None
        
    def containsPoint(self,point):
        if self.solveT(point) is not None:
            return True
        return False
        
    def compareEnd(self, other):
        return self.end.compareValue(other.end) 
        
    def length(self):
        return 0
        
    def touches(self,other):
        phi = (self.A * other.D)/(other.A * self.D)
        v = (self.B * other.D - other.B * self.D)/(other.A * self.D * 2)
        return self.compareValue(other) == 0 and ((v == 1) or (phi + v == 0))

    def contains(self,other):
        phi = (self.A * other.D)/(other.A * self.D)
        v = (self.B * other.D - other.B * self.D)/(other.A * self.D * 2)
        return self.compareValue(other) == 0 and (v <= 0 and (phi + v >= 1))

    def disjoint(self,other):
        phi = (self.A * other.D)/(other.A * self.D)
        v = (self.B * other.D - other.B * self.D)/(other.A * self.D * 2)
        return self.compareValue(other) == 0 and (v > 1 or (phi + v < 0))

    def align(self, other):
        phi = (self.A * other.D)/(other.A * self.D)
        v = (self.B * other.D - other.B * self.D)/(other.A * self.D * 2)
        return self.compareValue(other) == 0 and (v == 0 or (phi + v == 1))

    def reconstruct(self, phi, v):
        phi = coordinate(phi)
        v = coordinate(v)
        newA = self.a.scale(phi **2)
        newB = (self.a.scale(2*v).add(self.b)).scale(phi)
        newC = self.a.scale(v*v).add(self.b.scale(v)).add(self.c)
        newBezierCurve = bezierCurve2D(self.sort, newA, newB, newC, True)
        return newBezierCurve
        
    def createCurve(self, startPoint, endPoint):
        if self.start == startPoint:
            tS = coordinate(0)
        else:
            tS = self.solveT(startPoint)
        if self.end == endPoint:
            tE = coordinate(1)
        else:
            tE = self.solveT(endPoint)
        phi = tE - tS
        v = tS
        return self.reconstruct(phi, v)

    def common(self,other):
        if self.disjoint(other) or self.touches(other):
            return False
        phi = (self.A * other.D)/(other.A * self.D)
        v = (self.B * other.D - other.B * self.D)/(other.A * self.D * 2)

        nStart = self.start
        nEnd = self.end
        if phi+v > 1:
            nEnd = other.end
        if v < 0:
            nStart = other.start
        return self.createCurve(nStart, nEnd)

    def combine(self,other):
        if self.disjoint(other):
            return self
            
        phi = (self.A * other.D)/(other.A * self.D)
        v = (self.B * other.D - other.B * self.D)/(other.A * self.D * 2)

        nStart = self.start
        nEnd = self.end
        if phi+v < 1:
            nEnd = other.end
        if v > 0:
            nStart = other.start
        return self.createCurve(nStart, nEnd)
        
    def complement(self,other):
        result = []
            
        phi = (self.A * other.D)/(other.A * self.D)
        v = (self.B * other.D - other.B * self.D)/(other.A * self.D * 2)
        if v < 0 < phi+v:
            nStart = self.start
            nEnd = other.start
            new = self.createCurve(nStart, nEnd)
            if self.hasAttribute():
                new.addAttribute(self.attribute)
            new.key = self.key
            result.append(new)
        if v < 1 < phi+v:
            nStart = other.end
            nEnd = self.end
            new = self.createCurve(nStart, nEnd)
            if self.hasAttribute():
                new.addAttribute(self.attribute)
            new.key = self.key
            result.append(new)
        return result
        
    #!!!!!!!!!!!!!!!!!
    def intersectSegment(self, other):
        pass      

    def replicateInd(self, sort = 0):
        if sort == 0:
            sort = self.sort
        return self.__class__(sort, self.start, self.controlPoint, self.end)
        
    def transformInd(self, transformSet):
        affine = transformSet.retrieveTransform(affineTransform2D)
        if affine is None:
            p0 = self.start
            p1 = self.controlPoint    
            p2 = self.end    
            newCurve = self.duplicateInd()
        else:
            p0 = self.start.transform(affine)
            p1 = self.controlPoint.transform(affine)
            p2 = self.end.transform(affine)
            newCurve = self.__class__(self.sort, p0, p1, p2)
        
        transf = transformSet.retrieveTransform(graphTransform)
        if transf is not None:
            p0 = transf.apply(p0)
            p1 = transf.apply(p1)
            p2 = transf.apply(p2)
            newCurve = self.__class__(self.sort, p0, p1, p2)
        return newCurve
                
    def containsInd(self, ind):
        if not self.contains(ind):
            return False
        if self.sort.type == 'sort-attribute':
            if not self.hasAttribute():
                if ind.hasAttribute():
                    return False
                else:
                    return True
            if ind.hasAttribute():
                if ind.attribute.checkPartOf(self.attribute):
                    return True
                else:
                    return False
            else:
                return True
        return True
        
    @staticmethod
    def parse(reader, sort):
        if reader.token() != '<':
            raise ParsingError(reader,'expected a "<" at the beginning of creating a line segment')
        if reader.newToken() != '(': #bracket
            raise ParsingError(reader,'expected a "(" ')
        x,y = vector2D.readVector(reader)
        p0 = vector2D(x,y)
            
        reader.newToken() # bracket (current token would be ',')
        
        x,y = vector2D.readVector(reader)
        p1 = vector2D(x,y)
        
        reader.newToken() # bracket (current token would be ',')
        
        x,y = vector2D.readVector(reader)
        p2 = vector2D(x,y)
            
        ind = bezierCurve2D(sort, p0, p1, p2)

        if reader.token() != '>':
            raise ParsingError(reader, '">" expected to close the line segment')
            
        reader.newToken() #next token
        
        return ind
        
        

class bezierCurve3D(plane3D):
    
    type = 'ind-bezierCurve3D'
    
    def __init__(self, sort, p1, p2, p3, *args, **kwargs):
        super(bezierCurve3D, self).__init__(sort, [p1,p2,p3])
        p1 = self.localize(p1)
        p2 = self.localize(p2)
        p3 = self.localize(p3)
        
        self.a = p3.add(p1).subtract(p2.scale(2))
        self.b = (p2.subtract(p1)).scale(2)
        self.c = p1
        self.start = p1
        self.controlPoint = p2
        self.end = p3
        
        self.A = 4 * (self.a.x ** 2 + self.a.y ** 2)
        self.B = 4 * (self.a.x * self.b.x + self.a.y * self.b.y)
        self.C = self.b.x ** 2 + self.b.y ** 2
        self.D = 2 * (self.a.y * self.b.x - self.a.x * self.b.y)
        self.evaluateCurve()
        
    def vertex(self):
        t = -self.B/(2*self.A)
        return self.pos(t)
        
    def evaluateCurve(self):
        t = -self.B/(2*self.A)
        vg = self.vectGradient(t)
        if vg < v0 or (vg == v0 and self.vectGradient(t-1) < v0):
            self.start, self.end = self.end, self.start
            self.b = self.a.scale(-2).subtract(self.b)
            self.c = self.start
            self.B = 4 * (self.a.x * self.b.x + self.a.y * self.b.y)
            self.C = self.b.x ** 2 + self.b.y ** 2
            self.D = 2 * (self.a.y * self.b.x - self.a.x * self.b.y)
        
    def __str__(self):
        return '<'+str(self.position(self.start))+','+str(self.position(self.controlPoint))+','+str(self.position(self.end))+'>'
        
    def getValue(self):
        return str(self)

    def printVal(self):
        return '<'+str(self.position(self.start))+','+str(self.position(self.controlPoint))+','+str(self.position(self.end))+'>'

    def printEquation(self):
        return str(self.a)+'*t^2 + '+str(self.b)+'*t + '+str(self.c)

    def printRhino(self, displacement = 0):
        s = ''
        p1 = self.position(self.start)
        p2 = self.position(self.controlPoint)
        p3 = self.position(self.end)
        s += '_Curve '+str(p1.x + displacement)+','+str(p1.y)+','+str(p1.z)+'  '+str(p2.x + displacement)+','+str(p2.y)+','+str(p2.z)+'  '+str(p3.x + displacement)+','+str(p3.y)+','+str(p3.z)+' EnterEnd'
        return s
        
    def equals(self, other):
        if self.normal != other.normal or self.scalar != other.scalar:
            return False
        return self.a == other.a and self.b == other.b and self.c == other.c
        
    def isNil(self):
        return self.start.x == self.controlPoint.x == self.end.x and self.start.y == self.controlPoint.y == self.end.y
        
    def compareValue(self, other, params = None):
        c = self.normal.compareValue(other.normal)
        if c != 0:
            return c
            
        c = self.scalar.compareValue(other.scalar)
        if c != 0:
            return c
            
        if params is None:
            l = 1/(((other.A**(3/2)) * (self.D**2))/((self.A**(3/2)) * (other.D**2)))
            phi = (self.A * other.D)/(other.A * self.D)
            v = (self.B * other.D - other.B * self.D)/(other.A * self.D * 2)
        else:
            l, phi, v = params
        c = self.pos(0).compareValue(other.pos(v).scale(l))
        if c != 0:
            return c
        return self.pos(1).compareValue(other.pos(phi+v).scale(l))
        
        
    def compareInterval(self, other):
        c = self.normal.compareValue(other.normal)
        if c != 0:
            return c
        c = self.scalar.compareValue(other.scalar)
        if c != 0:
            return c
            
        l = 1/(((other.A**(3/2)) * (self.D**2))/((self.A**(3/2)) * (other.D**2)))
        phi = (self.A * other.D)/(other.A * self.D)
        v = (self.B * other.D - other.B * self.D)/(other.A * self.D * 2)
        c = self.compareValue(other, params = (l,phi,v))
        if c != 0:
            return c
        c = v.compareValue(c0)
        if c != 0:
            return c
        c = (phi+v).compareValue(c1)
        if c != 0:
            return c
        return 0
        
    def pos(self, t):
        tc = coordinate(t)
        return self.a.scale(tc**2).add(self.b.scale(tc)).add(self.c)
        
    def curvature(self, t):
        tc = coordinate(t)
        return self.D / ((self.A * (tc ** 2)+ self.B * tc + self.C)**(3/2))
                
    def gradient(self, t):
        dx = 2 * self.a.x * t + self.b.x
        dy = 2 * self.a.y * t + self.b.y
        if dx == 0:
            return None
        return dy/dx
        
    def vectGradient(self, t):
        dx = 2 * self.a.x * t + self.b.x
        dy = 2 * self.a.y * t + self.b.y
        return vector3D(dx, dy)
    """
    def vertexGradient(self):
        t = -self.B/(2*self.A)
        dx = 2 * self.a.x * t + self.b.x
        dy = 2 * self.a.y * t + self.b.y
        if dx == 0:
            return dy
        return dy / dx
    """
                
    def codescriptor(self):
        point = point3D(self.sort, self.position(self.vertex()))
        point.assoc = self
        return point
        
    def solveT(self, point):
        a = self.a.x
        b = self.b.x
        c = self.c.x - point.x
        
        if a == 0:
            x1 = -c/b
            x2 = None
        else:
            d = b*b-4*a*c # discriminant
            if d < 0:
                return None
            elif d == 0:
                x1 = (-b)/(2*a)
                x2 = None
            else:
                x1 = (-b+d.sqrt())/(2*a)
                x2 = (-b-d.sqrt())/(2*a)
            
        a = self.a.y
        b = self.b.y
        c = self.c.y - point.y
        if a == 0:
            y1 = -c/b
            y2 = None
        else:
            d = b**2-4*a*c
            
            if d < 0:
                return None
            elif d == 0:
                y1 = (-b)/(2*a)
                y2 = None
            else:
                y1 = (-b+d.sqrt())/(2*a)
                y2 = (-b-d.sqrt())/(2*a)
        x1r = round(x1.value,3)
        y1r = round(y1.value,3)
        if x1r == y1r:
            return x1
        if y2 is not None and x1r == round(y2.value,3):
            return x1
        if x2 is not None:
            x2r = round(x2.value,3)
            if x2r == y1r:
                return x2
            if y2 is not None and x2r == round(y2.value,3):
                return x2
        return None
        
    def containsPoint(self,point):
        if not super(bezierCurve3D, self).containsPoint(point):
            return False
        if self.solveT(point) is not None:
            return True
        return False
        
    def compareEnd(self, other):
        return self.end.compareValue(other.end) 
        
    def length(self):
        """
        !!!!!!!!
        """
        return 0
        
    def touches(self,other):
        if not super(bezierCurve3D, self).equals(other):
            return False
        phi = (self.A * other.D)/(other.A * self.D)
        v = (self.B * other.D - other.B * self.D)/(other.A * self.D * 2)
        return self.compareValue(other) == 0 and ((v == 1) or (phi + v == 0))

    def contains(self,other):
        if not super(bezierCurve3D, self).equals(other):
            return False
        phi = (self.A * other.D)/(other.A * self.D)
        v = (self.B * other.D - other.B * self.D)/(other.A * self.D * 2)
        return self.compareValue(other) == 0 and (v <= 0 and (phi + v >= 1))

    def disjoint(self,other):
        if not super(bezierCurve3D, self).equals(other):
            return False
        phi = (self.A * other.D)/(other.A * self.D)
        v = (self.B * other.D - other.B * self.D)/(other.A * self.D * 2)
        return self.compareValue(other) == 0 and (v > 1 or (phi + v < 0))

    def align(self, other):
        if not super(bezierCurve3D, self).equals(other):
            return False
        phi = (self.A * other.D)/(other.A * self.D)
        v = (self.B * other.D - other.B * self.D)/(other.A * self.D * 2)
        return self.compareValue(other) == 0 and (v == 0 or (phi + v == 1))

    def reconstruct(self, phi, v):
        phi = coordinate(phi)
        v = coordinate(v)
        newA = self.a.scale(phi **2)
        newB = (self.a.scale(2*v).add(self.b)).scale(phi)
        newC = self.a.scale(v*v).add(self.b.scale(v)).add(self.c)
        nCtrl = newB.scale(c0p5).add(newC)
        nEnd = newA.subtract(newC).add(nCtrl).add(nCtrl)
        newBezierCurve = bezierCurve3D(self.sort, self.position(newC), self.position(nCtrl), self.position(nEnd), True)
        return newBezierCurve
        
    def createCurve(self, startPoint, endPoint):
        if self.start == startPoint:
            tS = coordinate(0)
        else:
            tS = self.solveT(startPoint)
        if self.end == endPoint:
            tE = coordinate(1)
        else:
            tE = self.solveT(endPoint)
        phi = tE - tS
        v = tS
        return self.reconstruct(phi, v)

    def common(self,other):
        if self.disjoint(other) or self.touches(other):
            return False
        phi = (self.A * other.D)/(other.A * self.D)
        v = (self.B * other.D - other.B * self.D)/(other.A * self.D * 2)

        nStart = self.start
        nEnd = self.end
        if phi+v > 1:
            nEnd = other.end
        if v < 0:
            nStart = other.start
        return self.createCurve(nStart, nEnd)

    def combine(self,other):
        if self.disjoint(other):
            return self
            
        phi = (self.A * other.D)/(other.A * self.D)
        v = (self.B * other.D - other.B * self.D)/(other.A * self.D * 2)

        nStart = self.start
        nEnd = self.end
        if phi+v < 1:
            nEnd = other.end
        if v > 0:
            nStart = other.start
        return self.createCurve(nStart, nEnd)
        
    def complement(self,other):
        result = []
        phi = (self.A * other.D)/(other.A * self.D)
        v = (self.B * other.D - other.B * self.D)/(other.A * self.D * 2)
        if v < 0 < phi+v:
            nStart = self.start
            nEnd = other.start
            new = self.createCurve(nStart, nEnd)
            if self.hasAttribute():
                new.addAttribute(self.attribute)
            new.key = self.key
            result.append(new)
        if v < 1 < phi+v:
            nStart = other.end
            nEnd = self.end
            new = self.createCurve(nStart, nEnd)
            if self.hasAttribute():
                new.addAttribute(self.attribute)
            new.key = self.key
            result.append(new)
        return result
        
    #!!!!!!!!!!!!!!!!!
    def intersectSegment(self, other):
        pass      

    def replicateInd(self, sort = 0):
        if sort == 0:
            sort = self.sort
        return self.__class__(sort, self.position(self.start), self.position(self.controlPoint), self.position(self.end))
        
    def transformInd(self, transformSet):
        affine = transformSet.retrieveTransform(affineTransform3D)
        if affine is None:
            p0 = self.position(self.start)
            p1 = self.position(self.controlPoint)  
            p2 = self.position(self.end)
            newCurve = self.duplicateInd()
        else:
            p0 = self.position(self.start).transform(affine)
            p1 = self.position(self.controlPoint).transform(affine)
            p2 = self.position(self.end).transform(affine)
            newCurve = self.__class__(self.sort, p0, p1, p2)
        
        transf = transformSet.retrieveTransform(graphTransform)
        if transf is not None:
            p0 = transf.apply(p0)
            p1 = transf.apply(p1)
            p2 = transf.apply(p2)
            newCurve = self.__class__(self.sort, p0, p1, p2)
        return newCurve
                
    def containsInd(self, ind):
        if not self.contains(ind):
            return False
        if self.sort.type == 'sort-attribute':
            if not self.hasAttribute():
                if ind.hasAttribute():
                    return False
                else:
                    return True
            if ind.hasAttribute():
                if ind.attribute.checkPartOf(self.attribute):
                    return True
                else:
                    return False
            else:
                return True
        return True
        
    @staticmethod
    def parse(reader, sort):
        if reader.token() != '<':
            raise ParsingError(reader,'expected a "<" at the beginning of creating a line segment')
        if reader.newToken() != '(': #bracket
            raise ParsingError(reader,'expected a "(" ')
        x,y,z = vector3D.readVector(reader)
        p0 = vector3D(x,y,z)
            
        reader.newToken() # bracket (current token would be ',')
        
        x,y,z = vector3D.readVector(reader)
        p1 = vector3D(x,y,z)
        
        reader.newToken() # bracket (current token would be ',')
        
        x,y,z = vector3D.readVector(reader)
        p2 = vector3D(x,y,z)
            
        ind = bezierCurve3D(sort, p0, p1, p2)

        if reader.token() != '>':
            raise ParsingError(reader, '">" expected to close the line segment')
            
        reader.newToken() #next token
        
        return ind
        
        