# -*- coding: utf-8 -*-
"""
Created on Thu Jan  5 13:06:47 2017

@author: akibdpt
"""
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import range
from builtins import str

"""
from future import standard_library
standard_library.install_aliases()
"""
from .tree import tree
from .symbols import symbols
from .function import function
from ...io.parseReader import ParsingError, parseReader
from ...struct.coordinate import coordinate, sqrt, sin, cos, acos, asin, tan, isint, pi

class numberLiteral(tree):
    ABS = 'abs'
    SQRT = 'sqrt'
    SIN = 'sin'
    COS = 'cos'
    TAN = 'tan'
    ACOS = 'acos'
    ASIN = 'asin'
    SINDEG = 'sindeg'
    COSDEG = 'cosdeg'
    TANDEG = 'tandeg'
    ACOSDEG = 'acosdeg'
    ASINDEG = 'asindeg'
    FUNCTIONS = [ABS,SQRT,SIN,COS,ACOS,ASIN,TAN,SINDEG,COSDEG,TANDEG,ACOSDEG,ASINDEG]
    ASSISTS = [None,None,None,None,None,None,None,None,None,None,None,None]
    
    for fName in FUNCTIONS:
        function.reserve(fName)

        
    def __init__(self, value):
        self.value = value
        self.type = 'description-numberLiteral'
        
    def __str__(self):
        if isinstance(self.value, coordinate):
            if isint(self.value.value):
                return str(int(self.value.value))
        return str(self.value)
        
    def printVal(self):
        return 'numLit:' + str(self)
        
    def __eq__(self, other):
        if other.type != 'description-numberLiteral':
            return False
        return self.value == other.value
        
    def __len__(self):
        return len(str(self.value))
        
    def __lt__(self, other):
        return self.compareValue(other) == -1
        
    def __gt__(self, other):
        return self.compareValue(other) == 1
        
    def getNumber(self):
        return self.value
        
    def __hash__(self):
        hashNum = 5
        return  17 * hashNum + hash((self.value ** self.value).value)
        
    def apply(self, function, value = None):
        if function == self.ABS:
            return numberLiteral(abs(self.value))
        elif function == self.SQRT:
            return numberLiteral(coordinate(sqrt(self.value.value)))
        elif function == self.SIN:
            return numberLiteral(coordinate(sin(self.value.value)))
        elif function == self.COS:
            return numberLiteral(coordinate(cos(self.value.value)))
        elif function == self.TAN:
            return numberLiteral(coordinate(tan(self.value.value)))
        elif function == self.ASIN:
            return numberLiteral(coordinate(asin(self.value.value)))
        elif function == self.ACOS:
            return numberLiteral(coordinate(acos(self.value.value)))
        elif function == self.SINDEG:
            return numberLiteral(coordinate(sin((self.value*pi/180).value)))
        elif function == self.COSDEG:
            return numberLiteral(coordinate(cos((self.value*pi/180).value)))
        elif function == self.TANDEG:
            return numberLiteral(coordinate(tan((self.value*pi/180).value)))
        elif function == self.ASINDEG:
            return numberLiteral(coordinate(asin(self.value.value)*180/pi))
        elif function == self.ACOSDEG:
            return numberLiteral(coordinate(acos(self.value.value)*180/pi))
        raise ValueError('function name '+function+' not recognized')
        
    
    def compareValue(self, other):
        if isinstance(other,numberLiteral):
            c = self.value - other.value
        elif (other.type == 'description-numericExpression' or 
              other.type == 'description-literal' or 
              other.type == 'description-stringExpression'):
            return -other.compareValue(self)
        else:
            c = hash(self.__class__) - hash(other.__class__)
        if c > 0:
            return 1
        elif c < 0:
            return -1
        return 0
        
    def evaluate(self, assignment, sort):
        return self
        
    @staticmethod
    def functionExists(name, assist = None):
        for i in range(len(numberLiteral.FUNCTIONS)):
            if numberLiteral.FUNCTIONS[i]  == name and numberLiteral.ASSIST[i] is None:
                if assist is None:
                    if numberLiteral.ASSIST[i] is None:
                        return True
                elif numberLiteral.ASSIST[i] == assist:
                    return True
        return False
        
    @staticmethod
    def parse(desc, reader):
        if reader.token() != parseReader.NUMBER and reader.token() != parseReader.MINUS:
            raise ParsingError(reader, "expected a (positive or negative) number")
        if reader.token() == parseReader.MINUS and reader.previewToken() != parseReader.NUMBER:
            raise ParsingError(reader, "expected a number following the '-' sign")
        if reader.token() == parseReader.NUMBER:
            result = numberLiteral(coordinate(reader.value()))
        else:
            reader.newToken()
            result = numberLiteral(coordinate(reader.value())*(-1))
        reader.newToken()
        return result
        