# -*- coding: utf-8 -*-
"""
Created on Tue Jan 10 17:10:18 2017

@author: akibdpt
"""
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import str
"""
from future import standard_library
standard_library.install_aliases()
"""
from .tree import tree, DescriptionEvaluatingError
from .symbols import symbols
from ...io.parseReader import parseReader, ParsingError


from .literal import literal
from .tuple import tupleDesc

class parameter(tree):
    def __init__(self, name):
        if (literal.isLiteral(name)):
            raise ValueError(name + " already exists as a literal")
        self.type = 'description-parameter'
        self.name = name
        self.quantifier = 0
        self.condition = None
        
    def __str__(self):
        if self.quantifier != 0:
            return self.name + str(self.quantifier)
        if self.condition is not None:
            return self.name + symbols.CONDITIONAL + str(self.condition)
        return self.name
            
    def printVal(self):
        if self.quantifier != 0:
            return 'param:'+self.name + str(self.quantifier)
        if self.condition is not None:
            return 'param:'+self.name + symbols.CONDITIONAL + self.condition.printVal()
        return 'param:'+self.name
    
    def hasQuantifier(self):
        return self.quantifier != 0
        
    def hasCondition(self):
        return self.condition is not None
    
    def isEnumerated(self):
        return self.condition is not None and self.condition.isEnumerative()
        
    def matches(self, target, index):
        if not self.isEnumerated():
            return []
        return self.condition.match(target, index, self.condition)
        
    def __iter__(self):
        return iter([self])
        
    def __eq__(self, other):
        if not other.type == 'description-parameter':
            return False
        return self.name == other.name
        
    def getNumber(self, assignment, sort):
        num = self.evaluate(assignment, sort)
        return num.value
        
    def __lt__(self, other):
        return self.compareValue(other) == -1
        
    def __gt__(self, other):
        return self.compareValue(other) == 1
        
    def __hash__(self):
        hashNum = 3
        return 29 * hashNum + hash(self.name)
        
    def compareValue(self, other):
        if other.type == 'description-parameter':
            if self.name > other.name:
                return 1
            elif self.name < other.name:
                return -1
            this_suffix = self.quantifier
            other_suffix = other.quantifier
            if self.condition is not None:
                this_suffix = '?'
            if other.condition is not None:
                other_suffix = '?'
            if str(this_suffix) > str(other_suffix):
                return 1
            elif str(this_suffix) < str(other_suffix):
                return -1
            elif self.condition is not None:
                return self.condition.compareValue(other.condition)
            else:
                return 0
        elif (other.type == 'description-stringExpression' or 
                    other.type == 'description-numericExpression' or 
                    other.type == 'description-tupleExpression'):
            return - other.compareValue(self)
        c = hash(self.__class__) - hash(other.__class__)
        if c > 0:
            return 1
        if c < 0:
            return -1
        return 0
        

    def map(self, params, sort, onto):
        if onto is None:
            raise TypeError('description tree cannot be None')
        if self.condition is not None:
            if self.condition.matches(params, sort, onto):
                params.add(sort, self.name, onto)
                return True
            else:
                return False
        if (onto.type != 'description-literal' and 
            onto.type != 'description-numberLiteral' and
            onto.type != 'description-stringLiteral' and
            onto.type != 'description-tuple' and
            onto.type != 'description-kleeneTuple'):
                return False
        return params.add(sort, self.name, onto)
        
    def evaluate(self, assignment, sort):
        retr = assignment.retrieve(sort, self.name)
        if retr is None:
            raise DescriptionEvaluatingError('parameter name "'+self.name+'" of sort "'+sort.name+'" does not exists in the param set: '+str(assignment.parameters))
        if retr.type != 'description-kleeneTuple':
            result = retr.duplicate()
        else:
            result = retr
        return result
        
    def tupleExprEvaluate(self, assignment, sort):
        result = assignment.retrieve(sort, self.name).duplicate()
        if result.type != 'description-tuple':
            result = tupleDesc(symbols.PAREN_OPEN, symbols.COMMA, result)
        return result

    def similarFragment(self, other):
        if self.type != other.type:
            return False
        if self.name != other.name:
            return False
        if self.condition is None or other.condition is None or self.condition == other.condition:
            return True
        return False
        
    def mergeFragment(self, other):
        if self.condition is None:
            self.condition = other.condition

    @staticmethod
    def parse(desc, reader, qualified):
        if reader.token() != parseReader.IDENTIFIER:
            raise ParsingError(reader, "expected an identifier")
        name = reader.value()
        if desc.isKeyword(name):
            raise ParsingError(reader, "parameter name is a reserved keyword")
        result = parameter(name)
        if qualified:
            # if token == '*' and preview token is not number/identifier --> qualifier = *
            if (reader.newToken() == symbols.KLEENE_STAR and reader.previewToken() != parseReader.NUMBER and
                reader.previewToken() != parseReader.MINUS and reader.previewToken() != parseReader.IDENTIFIER):
                result.quantifier = reader.token()
                reader.newToken()
            # elif token == '?' --> condition
            elif reader.token() == symbols.CONDITIONAL:
                reader.newToken()
                result.condition = desc.condition.parse(desc, reader)
        else:
            reader.newToken()
        return result
        
    