# -*- coding: utf-8 -*-
"""
Created on Tue Aug 23 12:22:54 2016

@author: akibdpt
"""
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import absolute_import
from builtins import super
from builtins import str
"""
from future import standard_library
standard_library.install_aliases()
"""
from .individual import individual
from ..transform.affineTransform import affineTransform2D, affineTransform3D
from ..transform.graphTransform import graphTransform, graphTransform3D
from ..struct.vector import vector2D, vector3D
from ..struct.coordinate import c0
from ..io.parseReader import ParsingError

"""
 * A line is a linear, connected, non-bounded planar curve with a
 * discrete behavior.
 * 
 * The Line class extends on the Individual class.
 * It defines the characteristic individual for lines. A line is represented as
 * a direction vector and a position vector specifying the root of the line.
 * This characteristic individual accepts no parameters.
 * It specifies a {sortal.map.similarityMap} as default.
 * Forms of lines adhere to a discrete behavior.
"""
class line2D(individual):

    """
    initialize a line
    2 points (start and end) are passed into the function.
    the direction of the line will be calculated by having end subtracting start
                                (the direction's x value must always be >= 0)
    the unit direction is also calculated by dividing direction by its length
    root is the root vector of the line
    """
    type = 'ind-line2D'
    
    def __init__(self, sort, start, end, *args, **kwargs):
        super(line2D, self).__init__(sort)     

        if end.type == 'ind-point2D':
            end = end.value
        if start.type == 'ind-point2D':
            start = start.value
        if end.compareValue(start) == -1:
            temp = start
            start = end
            end = temp
        self.direction = (end.subtract(start))
        self.unitDir = self.direction.unit()
        self.root = start.subtract(self.direction.scale(start))
        
    def __str__(self):
        return '<*'+str(self.root)+','+str(self.root.add(self.direction))+'*>'
        return '<'+str(self.root)+','+str(self.root.add(self.direction))+'>'
                
    def getValue(self):
        return str(self)
        
    def printVal(self):
        return ('direction: '+str(self.direction)+ ', unitDir: '+
                str(self.unitDir) + ', root: '+str(self.root))
                
    def printRhino(self, displacement = 0):
        start = self.root.add(self.direction.scale(-10))
        end = self.root.add(self.direction.scale(10))
        return '_Line '+str(start.x + displacement)+','+str(start.y)+' '+str(end.x + displacement)+','+str(end.y)
        
    def isNil(self):
        return self.direction.x  == c0 and self.direction.y  == c0
        
    def equals(self, other):
        return (self.unitDir == other.unitDir) and (self.root == other.root)
        
    def __ne__(self, other):
        return not self == other
 
    def containsPoint(self,point):
        return point.subtract(self.root).parallel(self.direction)
        
    def compareValue(self, other):
        c = self.unitDir.compareValue(other.unitDir) 
        if c != 0:
            return c
        return self.root.compareValue(other.root) # c = self.root.compareValue(other.root)

    """
    check if two lines are parallel
    """
    def parallel(self, other):
        if other is None or 'ind-line' not in other.type :
            return False
        #return self.unitDir == other.unitDir and self.root != other.root
        return self.direction.parallel(other.direction)
        
    """
    check if two lines intersect
    """
    def intersect(self, other):
        if other is None or 'ind-line' not in other.type :
            return False
        return not self.direction.parallel(other.direction)
     
    """
    find the intersection of two lines
    """
    def intersection(self, other):
        if not self.intersect(other):
            print(self.printRhino(), other.printRhino(), self.intersect(other))
            raise ValueError('the two lines are not intersecting')
        if self.direction.x  == c0:
            a2 = other.direction.y / other.direction.x
            x = self.root.x
            b2 = other.root.y - a2*other.root.x
            y = a2*x+b2
        elif other.direction.x  == c0:
            a1 = self.direction.y / self.direction.x
            x = other.root.x
            b1 = self.root.y - a1*self.root.x
            y = a1*x+b1
        else:
            a1 = self.direction.y / self.direction.x
            a2 = other.direction.y / other.direction.x
            b1 = self.root.y - a1*self.root.x
            b2 = other.root.y - a2*other.root.x
            x = (b2-b1)/(a1-a2)
            y = a1*x+b1
        return vector2D(x,y)
        
    """
    check if two lines are perpendicular
    """
    def perpendicular(self,other):
        if other is None or 'ind-line' not in other.type :
            return False
        return self.direction.perpendicular(other.direction)
        
    def projection(self, point):
        v = point.subtract(self.root)
        return self.root.add(self.direction.scale(v))

    def replicateInd(self):
        return line2D(self.sort, self.root.add(self.direction), self.root)
        
        
    def transformInd(self, transformSet):
        affine = transformSet.retrieveTransform(affineTransform2D)
        if affine is None:
            return self.duplicateInd()
        start = self.root.transform(affine);
        end = self.root.add(self.direction).transform(affine);            
        return line2D(self.sort, start, end)
        
        
        
    @staticmethod
    def parse(reader, sort):
        if reader.token() != '<':
            raise ParsingError(reader,'expected a "<" at the beginning of creating a line segment')
        if reader.newToken() != '(': #bracket
            raise ParsingError(reader,'expected a "(" ')
        x,y = vector2D.readVector(reader)
        start = vector2D(x,y)
            
        reader.newToken() # bracket (current token would be ',')
        
        x,y = vector2D.readVector(reader)
        end = vector2D(x,y)
            
        ind = line2D(sort, start, end)

        if reader.token() != '>':
            raise ParsingError(reader, '">" expected to close the line segment')
            
        reader.newToken() #next token
        
        return ind
        
        

class line3D(individual):

    """
    initialize a line
    2 points (start and end) are passed into the function.
    the direction of the line will be calculated by having end subtracting start
                                (the direction's x value must always be >= 0)
    the unit direction is also calculated by dividing direction by its length
    root is the root vector of the line
    """
    type = 'ind-line3D'
    
    def __init__(self, sort, start, end, *args, **kwargs):
        super(line3D, self).__init__(sort)     

        if end.type == 'ind-point3D':
            end = end.value
        if start.type == 'ind-point3D':
            start = start.value
        if end.compareValue(start) == -1:
            temp = start
            start = end
            end = temp
        self.direction = end.subtract(start)
        self.unitDir = self.direction.unit()
        self.root = start.subtract(self.direction.scale(start))
        
    def __str__(self):
        return '<*'+str(self.root)+','+str(self.root.add(self.direction))+'*>'
                
    def getValue(self):
        return str(self)
        
    def printVal(self):
        return ('direction: '+str(self.direction)+ ', unitDir: '+
                str(self.unitDir) + ', root: '+str(self.root))
        
    def printRhino(self, displacement = 0):
        start = self.root.add(self.direction.scale(-10))
        end = self.root.add(self.direction.scale(10))
        return '_Line '+str(start.x + displacement)+','+str(start.y)+','+str(start.z)+' '+str(end.x + displacement)+','+str(end.y)+','+str(end.z)
        
    def isNil(self):
        return self.direction.x  == c0 and self.direction.y  == c0 and self.direction.z  == c0
        
    def equals(self, other):
        return (self.unitDir == other.unitDir) and (self.root == other.root)
            
    def __ne__(self, other):
        return not self == other
        
    def compareValue(self, other):
        c = self.unitDir.compareValue(other.unitDir) 
        if c != 0:
            return c
        return self.root.compareValue(other.root) # c = self.root.compareValue(other.root)
        
    def containsPoint(self,point):
        return point.subtract(self.root).parallel(self.direction)
        
    def projection(self, point):
        v = point.subtract(self.root)
        return self.root.add(self.direction.scale(v))

    """
    check if two lines are parallel
    """
    def parallel(self, other):
        if other is None or 'ind-line' not in other.type :
            return False
        return self.direction.parallel(other.direction)
        
    """
    check if two lines are coPlanar
    """
    def coPlanar(self, other):
        if other is None or 'ind-line' not in other.type :
            return False
        return self.direction.product(other.direction).dotproduct(self.root.subtract(other.root))  == c0
        
    """
    check if two lines are skewed
    """
    def skew(self, other):
        if other is None or 'ind-line' not in other.type :
            return False
        return not self.root.subtract(other.root).dotproduct(self.direction.product(other.direction))  == c0
        
    """
    check if two lines intersect
    """
    def intersect(self, other):
        if other is None or 'ind-line' not in other.type :
            return False
        return self.coPlanar(other) and not self.parallel(other)
        
        
    """
    check if two lines intersect
    """
    def perpendicular(self, other):
        if other is None or 'ind-line' not in other.type :
            return False
        return self.direction.dotproduct(other.direction)  == c0
     
    """
    find the intersection of two lines
    """
    def intersection(self, other):
        if not self.intersect(other):
            raise ValueError('the two lines are not intersecting')
        denom = self.unitDir.product(other.unitDir)
        diff = other.root.subtract(self.root)
        numer1 = diff.product(other.unitDir)
        
        if denom.x != 0:
            x = numer1.x/denom.x
        elif denom.y != 0:
            x = numer1.y/denom.y
        else:
            x = numer1.z/denom.z
        return self.root.add(self.unitDir.scale(x))
                
    def replicateInd(self):
        return line3D(self.sort, self.root.add(self.direction), self.root)
        
    def transformInd(self, transformSet):
        affine = transformSet.retrieveTransform(affineTransform3D)
        if affine is None:
            newLine = self.duplicateInd()
        else:
            start = self.root.transform(affine);
            end = self.root.add(self.direction).transform(affine);     
            newLine = line3D(self.sort, start, end)
        transf = transformSet.retrieveTransform(graphTransform3D)
        if transf is not None:
            vects = []
            for i in transf.vects:
                if self.containsPoint(i[0]):
                    vects.append(i[1])
                if len(vects) >=2:
                    break
            newLine = self.__class__(self.sort, vects[0], vects[1])
        return newLine
        
        
        
    @staticmethod
    def parse(reader, sort):
        if reader.token() != '<':
            raise ParsingError(reader,'expected a "<" at the beginning of creating a line segment')
        if reader.newToken() != '(': #bracket
            raise ParsingError(reader,'expected a "(" ')
        x,y,z = vector3D.readVector(reader)
        start = vector3D(x,y,z)
            
        reader.newToken() # bracket (current token would be ',')
        
        x,y,z = vector3D.readVector(reader)
        end = vector3D(x,y,z)
            
        ind = line3D(sort, start, end)

        if reader.token() != '>':
            raise ParsingError(reader, '">" expected to close the line segment')
            
        reader.newToken() #next token
        
        return ind
