# -*- coding: utf-8 -*-
"""
Created on Tue Sep  6 12:37:59 2016

@author: akibdpt
"""
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import absolute_import
from builtins import round
from builtins import range
"""
from future import standard_library
standard_library.install_aliases()
"""
from .mapping import mapping
from .distinguishable import distinguishable
from ..struct.coordinate import coordinate, c0p5, c2
from ..transform.transformSet import transformSet
from ..transform.affineTransform import affineTransform3D
from ..transform.transform import transform
from ..ind.plane import plane3D
from ..ind.line import line3D
from ..ind.ellipse import ellipse3D
from ..ind.circle import circle3D
from ..ind.point import point3D
from enum import Enum

p = coordinate.p
arcClasses = (circle3D, ellipse3D)

class similarityMap3D(mapping):
    
    type = 'map-distinguishable-similarityMap3D'

    class casus(Enum):
        CASE_3PT_NOT_COLINEAR = [1,3],  """DONE"""
        CASE_2PT_1PL_NOT_PERPENDICULAR = [2,3], """DONE"""
        
        CASE_1PT_1ARC_NOT_COINCIDENT = [3,2], """DONE"""
        
        CASE_1PT_1LN_NOT_COLINEAR = [4,2], """DONE"""
        CASE_1PT_1LN_1PL_NOT_COINCIDENT_NOT_PERPENDICULAR = [5,3],  #***
        CASE_1PT_2PL_NOT_CONCURRENT = [6,3], #***
        
        CASE_2ARC_NOT_CONCENTRIC = [7,2], """DONE"""
        CASE_2ARC_NOT_PARALLEL = [8,2], #********#
        CASE_1ARC_1LN_NOT_COLINEAR = [9,2], #********#
        
        CASE_3LN_NO_TWO_PARALLEL = [10,3], """DONE"""
        CASE_3LN_FIRST_TWO_PARALLEL = [11,3], """DONE"""
        CASE_3LN_LAST_TWO_PARALLEL = [12,3], """DONE"""
        CASE_2LN_SKEWED = [13,2], """DONE"""
        
        CASE_2LN_1PL_NOT_PARALLEL_NOR_CONCURRENT = [14,3], """DONE"""
        CASE_2LN_1PL_PARALLEL_FIRST = [15,3], """DONE"""
        CASE_2LN_1PL_PARALLEL_BOTH = [16,3], """DONE"""
        
        CASE_1LN_2PL_NOT_ALL_CONCURRENT = [17,3], """DONE"""
        CASE_4PL_NOT_ALL_PARALLEL = [18,4], #***
        REDUNDANT = 0
        
        

    class distinguishes(Enum):
        INSUFFICIENT = 1, 
        UNSUITABLE = 2, 
        SUCCESS = 3

    def __init__(self, *args, **kwargs):
        super(similarityMap3D, self).__init__()     
        self.distinguishables = []
        self.transforms = []
        self.fromSet = [None,None,None]
        self.checkTree = None
        self.distinguishableSuccess = False
        self.reflection = 0
        self.transformset = transformSet()
        self.casus = similarityMap3D.casus.REDUNDANT
        self.subCase = []
        self.cosAngle = None
        self.ratio = None
        return
        
    @staticmethod
    def getInstance():
        return similarityMap3D()
    
    def clear(self):
        for i in self.distinguishables:
            i.clear()
        self.transforms = []
        return
        
    def isValid(self):
        if self.distinguishableSuccess:
            return True
        if len(self.distinguishables) == 1:
            if (self.distinguishables[0].element.individual.type == 'ind-point3D' 
            or self.distinguishables[0].element.individual.type == 'ind-line3D'):
                return True
        return False
    
    def map(self, fromInd, toInd):
        for i in self.distinguishables:
            elmInd = i.element.individual
            if elmInd == fromInd:
                if elmInd.hasAttribute():
                    if fromInd.hasAttribute() and elmInd.attribute.checkPartOf(toInd.attribute):
                        i.addCounterpart(toInd)
                        return True
                else:
                    i.addCounterpart(toInd)
                    return True
        return False
    
    def distinguish(self, ind):
        element = ind.codescriptor()
        if ind.hasAttribute():
            element.addAttribute(ind.attribute)
        dist = None
        for i in self.distinguishables:
            if element == i.element.individual:
                dist = i
                dist.addInstance(ind)
                return dist
        if self.distinguishableSuccess:
            return None
        dist = distinguishable(element)
        
        #cell =  [element, dist]    
        self.distinguishables.append(dist)
        dist.addInstance(ind)
        size = len(self.distinguishables)
        last = element
        first = None
        second = None
        third = None
        if size > 1:
            first = self.distinguishables[0].element.individual
            if size > 2:
                second = self.distinguishables[1].element.individual
                if size > 3:
                    third = self.distinguishables[2].element.individual
        success = self.assessDistinguishables(size, last, first, second, third)
        if success == similarityMap3D.distinguishes.UNSUITABLE:
            del self.distinguishables[-1]
        if success == similarityMap3D.distinguishes.SUCCESS:
            self.distinguishableSuccess = True
        return dist

    def assessDistinguishables(self, size, last, first, second, third):
        if size == 1:
            if last.__class__ in (point3D, line3D, plane3D, circle3D, ellipse3D):
                return similarityMap3D.distinguishes.INSUFFICIENT
        if size == 2 : 
            #cases with 2 lines
            if last.__class__ == line3D and first.__class__ == line3D:
                if last.skew(first):
                    return self.assessSkew(first, last)
                return similarityMap3D.distinguishes.INSUFFICIENT

            #cases with 2 points
            elif last.__class__ == point3D and first.__class__ == point3D:
                if last == first:
                    return similarityMap3D.distinguishes.UNSUITABLE
                return similarityMap3D.distinguishes.INSUFFICIENT

            #cases with 1 point + 1 plane
            elif last.__class__ == plane3D and first.__class__ == point3D:
                if last.containsPoint(first.value):
                    return similarityMap3D.distinguishes.UNSUITABLE
                return similarityMap3D.distinguishes.INSUFFICIENT
                
            #cases with 1 line + 1 plane
            elif last.__class__ == plane3D and first.__class__ == line3D:
                if last.containsLine(first):
                    return similarityMap3D.distinguishes.UNSUITABLE
                return similarityMap3D.distinguishes.INSUFFICIENT
                
        if size >= 2:
            
            
            #*********************************************************            
            #case: 1 point 1 arc not coincident
            if first.__class__ == point3D and last.__class__ in arcClasses:
                lastCenter = last.position(last.center)
                
                point = first.value
                if first.value == lastCenter:
                    if second is None or second.value == lastCenter:
                        return similarityMap3D.distinguishes.UNSUITABLE
                    point = second.value
                    self.distinguishables.pop(0)
                elif second is not None:
                    self.distinguishables.pop(1)
                                        
                self.fromSet[0] = point
                self.fromSet[1] = lastCenter
                
                u = last.normal
                v = point.subtract(lastCenter)
                
                if u.parallel(v):
                    return similarityMap3D.distinguishes.UNSUITABLE
                                    
                self.fromSet[2] = self.fromSet[1].add(u.scale(((v.dotproduct(v)/u.dotproduct(u))**c0p5)*c2))
                
                self.cosAngle = u.unit().dotproduct(v.unit())
                self.subCase.append(last.__class__)
                if self.cosAngle == 0:
                    self.subCase.append(last.assoc.containsPoint(self.fromSet[0]))
                else:
                    self.subCase.append(False)
                                
                self.casus = similarityMap3D.casus.CASE_1PT_1ARC_NOT_COINCIDENT
                return similarityMap3D.distinguishes.SUCCESS         
                
            #case: 1 point + 1 line not colinear
            if first.__class__ == point3D and last.__class__ == line3D:
                self.fromSet[1] = first.value
                u = last.direction
                v = self.fromSet[1].subtract(last.root)
                
                if u.parallel(v):
                    if second is None:
                        return similarityMap3D.distinguishes.UNSUITABLE
                    self.fromSet[1] = second.value
                    v = self.fromSet[1].subtract(last.root)
                    if u.parallel(v):
                        return similarityMap3D.distinguishes.UNSUITABLE
                    self.distinguishables.pop(0)
                elif second is not None:
                    self.distinguishables.pop(1)
                    
                self.fromSet[0] = last.root.add(u.scale(v))
                self.subCase.append(last.assoc.containsPoint(self.fromSet[0]))
                self.subCase.append(not(last.assoc.start == self.fromSet[0] or last.assoc.end == self.fromSet[0]))
                
                d = self.fromSet[1].subtract(self.fromSet[0])
                self.fromSet[2] = self.fromSet[0].add(u.scale(((d.dotproduct(d)/u.dotproduct(u))**c0p5)*c2))

                if not last.assoc.containsPoint(self.fromSet[2]):
                    self.fromSet[2] = self.fromSet[0].subtract(u.scale(((d.dotproduct(d)/u.dotproduct(u))**c0p5)*c2))
                
                
                self.subCase.append(last.assoc.containsPoint(self.fromSet[2]))
                
                self.casus = similarityMap3D.casus.CASE_1PT_1LN_NOT_COLINEAR
                return similarityMap3D.distinguishes.SUCCESS         
                
            #case: 2 arcs not concentric / 2 concentric arcs with normals not parallel
            if first.__class__ in arcClasses and last.__class__ in arcClasses:
                firstCenter = first.position(first.center)
                lastCenter = last.position(last.center)
                
                #case: 2 arcs not concentric
                if firstCenter != lastCenter:
                    firstArc = first
                    lastArc = last
                    self.fromSet[0] = firstCenter
                    self.fromSet[1] = lastCenter
                    u = lastArc.normal
                    v = firstCenter.subtract(lastCenter)
                    
                    if u.parallel(v):
                        firstArc = last
                        lastArc = first
                        self.fromSet[0] = lastCenter
                        self.fromSet[1] = firstCenter
                        u = lastArc.normal
                        v = lastCenter.subtract(firstCenter)
                        if u.parallel(v):
                            return similarityMap3D.distinguishes.UNSUITABLE
                            
                    self.fromSet[2] = self.fromSet[1].add(u.scale(((v.dotproduct(v)/u.dotproduct(u))**c0p5)*c2))
                    
                    self.cosAngle = u.unit().dotproduct(v.unit())

                    distanceVec = self.fromSet[1].subtract(self.fromSet[0])
                    distanceSqr = distanceVec.dotproduct(distanceVec)

                    if firstArc.__class__ == circle3D:
                        firstRadius = firstArc.radius
                    else:
                        firstRadius = firstArc.radius[0]
                    
                    if lastArc.__class__ == circle3D:
                        lastRadius = lastArc.radius
                    else:
                        lastRadius = lastArc.radius[0]
                    
                    self.ratio = [firstRadius/lastRadius, (firstRadius**2) / distanceSqr]
                    
                    self.subCase.append(firstArc.__class__)
                    self.subCase.append(lastArc.__class__)
                                        
                    self.casus = similarityMap3D.casus.CASE_2ARC_NOT_CONCENTRIC
                    return similarityMap3D.distinguishes.SUCCESS
                    
                #case: 2 concentric arcs with normals not parallel
                elif not first.normal.parallel(last.normal):
                    self.fromSet[0] = firstCenter
                    
                    if first.__class__ == circle3D:
                        firstRadius = first.radius
                    else:
                        firstRadius = first.radius[0]
                    
                    if last.__class__ == circle3D:
                        lastRadius = last.radius
                    else:
                        lastRadius = last.radius[0]
                        
                    self.ratio = firstRadius/lastRadius
                        
                    self.subCase.append(first.__class__)
                    self.subCase.append(last.__class__)
                    
                    firstUnit = first.normal.unit()
                    lastUnit = last.normal.unit()
                    
                    self.cosAngle = firstUnit.dotproduct(lastUnit)
                    
                    self.fromSet[1] = self.fromSet[0].add(firstUnit.scale(firstRadius))
                    self.fromSet[2] = self.fromSet[0].add(lastUnit.scale(lastRadius))
                    
                    #self.checkTree = True
                    
                    self.casus = similarityMap3D.casus.CASE_2ARC_NOT_PARALLEL
                    return similarityMap3D.distinguishes.SUCCESS       
                else:
                    return similarityMap3D.distinguishes.UNSUITABLE
                    
                    
            #case: 1 arc + 1 line not colinear                    
            elif first.__class__ in arcClasses and last.__class__ == line3D:
                self.fromSet[1] = first.position(first.center)
                u = last.direction
                v = self.fromSet[1].subtract(last.root)
                
                if u.parallel(v):
                    if second is None:
                        return similarityMap3D.distinguishes.UNSUITABLE
                    self.fromSet[1] = second.center
                    v = self.fromSet[1].subtract(last.root)
                    if u.parallel(v):
                        return similarityMap3D.distinguishes.UNSUITABLE
                    self.distinguishables.pop(0)
                elif second is not None:
                    self.distinguishables.pop(1)
                    
                self.fromSet[0] = last.root.add(u.scale(v))
                self.subCase.append(last.assoc.containsPoint(self.fromSet[0]))
                self.subCase.append(not(last.assoc.start == self.fromSet[0] or last.assoc.end == self.fromSet[0]))
                
                d = self.fromSet[1].subtract(self.fromSet[0])
                self.fromSet[2] = self.fromSet[0].add(u.scale((d.dotproduct(d)/u.dotproduct(u))**c0p5*c2))

                if not last.assoc.containsPoint(self.fromSet[2]):
                    self.fromSet[2] = self.fromSet[0].subtract(u.scale((d.dotproduct(d)/u.dotproduct(u))**c0p5*c2))
                
                
                self.subCase.append(last.assoc.containsPoint(self.fromSet[2]))
                
                self.casus = similarityMap3D.casus.CASE_1ARC_1LN_NOT_COLINEAR
                return similarityMap3D.distinguishes.SUCCESS         
        
            #*********************************************************
                
            #*********************************************************
            #case: 3 points not coincident
            elif last.__class__ == point3D and first.__class__ == point3D and second.__class__ == point3D:
                if last == first or last == second:
                    return similarityMap3D.distinguishes.UNSUITABLE
                if last.value.colinear(first.value, second.value):
                    return similarityMap3D.distinguishes.UNSUITABLE
                    
                self.fromSet[0] = first.value      
                self.fromSet[1] = second.value                
                self.fromSet[2] = last.value
                
                self.casus = similarityMap3D.casus.CASE_3PT_NOT_COLINEAR
                return similarityMap3D.distinguishes.SUCCESS
            
            #cases with 3 lines
            elif last.__class__ == line3D and first.__class__ == line3D and second.__class__ == line3D:
                if last.skew(first):
                    del self.distinguishables[1]
                    return self.assessSkew(first, last)
                if last.skew(second):
                    del self.distinguishables[0]
                    return self.assessSkew(second, last)
                    
                fpl = last.parallel(first)
                fps = second.parallel(first)
                spl = second.parallel(last)
                if not (fpl or fps or spl):
                    self.fromSet[0] = first.intersection(last)                        
                    self.fromSet[1] = first.intersection(second)
                    self.fromSet[2] = second.intersection(last)
                        
                    #if self.fromSet[0] == self.fromSet[1] or self.fromSet[0] == self.fromSet[2] or self.fromSet[2] == self.fromSet[1]:
                    if self.fromSet[0] == self.fromSet[1] == self.fromSet[2]:
                        return similarityMap3D.distinguishes.UNSUITABLE
                        
                    self.subCase.append(first.assoc.containsPoint(self.fromSet[0]))
                    self.subCase.append(first.assoc.containsPoint(self.fromSet[1]))
                    self.subCase.append(second.assoc.containsPoint(self.fromSet[1]))
                    self.subCase.append(second.assoc.containsPoint(self.fromSet[2]))
                    self.subCase.append(last.assoc.containsPoint(self.fromSet[0]))
                    self.subCase.append(last.assoc.containsPoint(self.fromSet[2]))
                    
                    u = self.fromSet[1].subtract(self.fromSet[0])
                    v = self.fromSet[2].subtract(self.fromSet[1])
                    w = self.fromSet[2].subtract(self.fromSet[0])
                    
                    self.ratio = [u.squareLength(),v.squareLength(),w.squareLength()]
                    
                    #self.checkTree = True                    
                    
                    self.casus = similarityMap3D.casus.CASE_3LN_NO_TWO_PARALLEL
                    return similarityMap3D.distinguishes.SUCCESS

                if fpl and fps and spl:
                    return similarityMap3D.distinguishes.UNSUITABLE
                    
                if fpl:
                    self.distinguishables[1], self.distinguishables[-1] = self.distinguishables[-1], self.distinguishables[1]
                    second, last = last, second
                    fps = second.parallel(first)
                    
                if spl:
                    self.distinguishables[0], self.distinguishables[-1] = self.distinguishables[-1], self.distinguishables[0]
                    first, last = last, first
                    fps = True
                    
                #<<<<<<<<<<<<<<<<<<<<################>>>>>>>>>>>>>>>>>>>>>>>
                #<<<<<<<<<<<<<<<<<<<<################>>>>>>>>>>>>>>>>>>>>>>>
                #<<<<<<<<<<<<<<<<<<<<################>>>>>>>>>>>>>>>>>>>>>>>
                #<<<<<<<<<<<<<<<<<<<<################>>>>>>>>>>>>>>>>>>>>>>>
                if fps:
                    try:
                        self.fromSet[0] = first.intersection(last)
                        self.fromSet[1] = second.intersection(last)
                    except ValueError:
                        return similarityMap3D.distinguishes.UNSUITABLE
                        
                    u = first.direction
                    d = self.fromSet[1].subtract(self.fromSet[0])
                    v = u.scale((d.dotproduct(d)/(u.dotproduct(u)))**c0p5*c2)
                    self.fromSet[2] = self.fromSet[0].add(v)
                    if not first.assoc.containsPoint(self.fromSet[2]):
                        self.fromSet[2] = self.fromSet[0].subtract(v)
                    
                    self.subCase.append(first.assoc.containsPoint(self.fromSet[0]))
                    self.subCase.append(second.assoc.containsPoint(self.fromSet[1]))
                    self.subCase.append(last.assoc.containsPoint(self.fromSet[0]))
                    self.subCase.append(last.assoc.containsPoint(self.fromSet[1]))
                    self.subCase.append(first.assoc.containsPoint(self.fromSet[2]))
                    
                    u = first.unitDir
                    v = last.unitDir
                    self.cosAngle = abs(u.dotproduct(v))
                    
                    #self.checkTree = True       
                
                    self.casus = similarityMap3D.casus.CASE_3LN_FIRST_TWO_PARALLEL
                    return similarityMap3D.distinguishes.SUCCESS
            #*********************************************************
            #cases: 2 points, 1 plane
            elif last.__class__ == plane3D and first.__class__ == point3D and second.__class__ == point3D:
                foot = last.projection(first.value)
                if foot == first:
                    foot = last.projection(second.value)
                    if foot == second:
                        d = second.subtract(first)
                        u = last.normal.product(d)
                        foot = first.add(u.scale((d.dotProduct(d) / (u.dotProduct(u)))**c0p5*c2))
                        self.subCase = 0
                    else:
                        self.subCase = 2
                else:
                    self.subCase = 1

                if (foot.colinear(second.value, first.value)):
                    return similarityMap3D.distinguishes.UNSUITABLE
                self.fromSet[0] = first.value
                self.fromSet[1] = second.value
                self.fromSet[2] = foot
                self.casus = similarityMap3D.casus.CASE_2PT_1PL_NOT_PERPENDICULAR
                return similarityMap3D.distinguishes.SUCCESS
            
            #cases: 1 points, 2 plane
            elif last.__class__ == plane3D and first.__class__ == point3D and second.__class__ == plane3D:
                if last.parallel(second):
                    return similarityMap3D.distinguishes.UNSUITABLE
                self.fromSet[0] = first.value
                self.fromSet[1] = second.projection(first.value)
                self.fromSet[2] = last.projection(first.value)
                self.casus = similarityMap3D.casus.CASE_1PT_2PL_NOT_CONCURRENT
                return similarityMap3D.distinguishes.SUCCESS
            
            #cases: 2 lines, 1 plane
            elif last.__class__ == plane3D and first.__class__ == line3D and second.__class__ == line3D:
                if last.containsLine(first) or last.containsLine(second):
                    return similarityMap3D.distinguishes.UNSUITABLE
                if last.parallelLine(first) and last.parallelLine(second):
                    if first.parallel(second):
                        return similarityMap3D.distinguishes.UNSUITABLE
                    self.fromSet[0] = first.intersection(second)
                    self.fromSet[1] = last.projection(self.fromSet[0])
                    d0 = last.normal.product(first.unitDir)
                    l0 = self.fromSet[1].subtract(self.fromSet[0]).length()
                    self.fromSet[2] = self.fromSet[1].add(d0.scale(l0))
                    self.casus = similarityMap3D.casus.CASE_2LN_1PL_PARALLEL_BOTH
                    return similarityMap3D.distinguishes.SUCCESS
                if last.parallelLine(first) or last.parallelLine(second):
                    if last.parallelLine(second):
                        first, second = second, first
                    self.fromSet[0] = last.intersectionLine(second)
                    self.fromSet[1] = first.intersection(second)
                    l0 = self.fromSet[1].subtract(self.fromSet[0]).length()
                    self.fromSet[2] = self.fromSet[1].add(first.unitDir.scale(l0))
                    self.casus = similarityMap3D.casus.CASE_2LN_1PL_PARALLEL_FIRST
                    return similarityMap3D.distinguishes.SUCCESS
                self.fromSet[0] = last.intersectionLine(first)
                self.fromSet[1] = last.intersectionLine(second)
                if self.fromSet[0] == self.fromSet[1]:
                    return similarityMap3D.distinguishes.UNSUITABLE
                self.fromSet[2] = self.fromSet[0].add(last.normal.scale(self.fromSet[0].subtract(self.fromSet[1]).length()))
                self.casus = similarityMap3D.casus.CASE_2LN_1PL_NOT_PARALLEL_NOR_CONCURRENT
                return similarityMap3D.distinguishes.SUCCESS
                
            #cases: 1 line, 2 planes
            elif last.__class__ == plane3D and first.__class__ == line3D and second.__class__ == plane3D:
                if last.parallelLine(first) and second.parallelLine(first):
                    return similarityMap3D.distinguishes.UNSUITABLE
                if second.parallelLine(first):
                    last, second = second, last
                self.fromSet[0] = second.intersectionLine(first)
                self.fromSet[1] = last.projection(self.fromSet[0])
                if self.fromSet[0] == self.fromSet[1]:
                    return similarityMap3D.distinguishes.UNSUITABLE
                l0 = self.fromSet[1].subtract(self.fromSet[0]).length()
                self.fromSet[2] = self.fromSet[0].add(first.unitDir.scale(l0))
                
                self.subCase = True if last.parallelLine(first) else False                
                self.casus = similarityMap3D.casus.CASE_1LN_2PL_NOT_ALL_CONCURRENT
                return similarityMap3D.distinguishes.SUCCESS
            #*********************************************************
                    
        return similarityMap3D.distinguishes.UNSUITABLE

    def assessSkew(self, first, last):
        n = last.direction.product(first.direction)
        p1 = first.root
        p2 = last.root
        d1 = first.direction
        d2 = last.direction
        n1 = d1.product(n)
        n2 = d2.product(n)
        self.fromSet[0] = p1.add(d1.scale((p2.subtract(p1).dotproduct(n2))/(d1.dotproduct(n2))))
        self.fromSet[1] = p2.add(d2.scale((p1.subtract(p2).dotproduct(n1))/(d2.dotproduct(n1))))
        
        
        self.subCase.append(first.assoc.containsPoint(self.fromSet[0]))
        self.subCase.append(last.assoc.containsPoint(self.fromSet[1]))
        
        d = self.fromSet[1].subtract(self.fromSet[0])
        v = d1.scale((d.dotproduct(d)/(d1.dotproduct(d1)))**c0p5*c2)
        self.fromSet[2] = self.fromSet[0].add(v)
        check = first.assoc.containsPoint(self.fromSet[2])
        
        if check:
            self.subCase.append(check)
        else:
            self.fromSet[2] = self.fromSet[0].subtract(v)
            self.subCase.append(first.assoc.containsPoint(self.fromSet[2]))
            
        #self.checkTree = True
        
        self.casus = similarityMap3D.casus.CASE_2LN_SKEWED
        return similarityMap3D.distinguishes.SUCCESS


    def assessTargetSet(self, target):
        if self.casus == similarityMap3D.casus.CASE_3PT_NOT_COLINEAR:
            first = target[0].individual
            second = target[1].individual
            if first == second:
                return False
            return True
        elif self.casus == similarityMap3D.casus.CASE_2PT_1PL_NOT_PERPENDICULAR:
            first = target[0].individual
            second = target[1].individual
            if first.type == second.type and first == second:
                return False
        elif self.casus == similarityMap3D.casus.CASE_3LN_NO_TWO_PARALLEL:
            first = target[0].individual
            second = target[1].individual
            if first.parallel(second):
                return False
            return True
        elif self.casus == similarityMap3D.casus.CASE_3LN_FIRST_TWO_PARALLEL:
            first = target[0].individual
            second = target[1].individual
            if not first.parallel(second):
                return False
            if second.compareValue(first) == 0:
                return False
            return True
        elif self.casus == similarityMap3D.casus.CASE_3LN_LAST_TWO_PARALLEL:
            first = target[0].individual
            second = target[1].individual
            if first.parallel(second):
                return False
            return True
        elif self.casus == similarityMap3D.casus.CASE_2LN_1PL_PARALLEL_BOTH:
            first = target[0].individual
            second = target[1].individual
            if first.parallel(second):
                return False
            return True
        elif self.casus == similarityMap3D.casus.CASE_1LN_2PL_NOT_ALL_CONCURRENT:
            first = target[0].individual
            second = target[1].individual
            if second.parallelLine(first):
                return False
        return True
        
    def potentialTransforms(self): ########################
        #print('CASUS:', self.casus)
        if len(self.transforms) != 0:
            return self.transforms
            
        if self.casus == similarityMap3D.casus.REDUNDANT:
            return self.specialPotentialTransforms()
            
        size = len(self.distinguishables)
        source = []
        counterparts = [] 
                        
        for elm in self.distinguishables:
            source.append(elm.element)
            counterparts.append(elm.getCounterparts())
        target = [None]*size
        """
        if self.checkTree and size == 2:
            for i in counterparts:
                if len(i) == 0:
                    return self.transforms
            if counterparts[0][0].individual.sort.name == counterparts[1][0].individual.sort.name:
                self.checkTree = [[True for i in range(len(counterparts[1]))] 
                                        for j in range(len(counterparts[0]))]
            else:
                self.checkTree = True
                                     
        elif self.checkTree and size == 3:
            for i in counterparts:
                if len(i) == 0:
                    return self.transforms
            if (counterparts[0][0].individual.sort.name == 
                counterparts[1][0].individual.sort.name == 
                counterparts[2][0].individual.sort.name):
                self.checkTree = [[[True for i in range(len(counterparts[2]))] 
                                         for j in range(len(counterparts[1]))] 
                                         for k in range(len(counterparts[0]))]
            else:
                self.checkTree = True                                
        """
        """
        ##### TESTING: PRINTING ALL COUNTERPARTS #####
        print()
        for e0 in range(len(counterparts)):
            print('\ncounterpart set '+str(e0)+':')
            for e1 in range(len(counterparts[e0])):
                print(counterparts[e0][e1].individual.printVal())
        """
        for e0 in range(len(counterparts[0])):
            target[0] = counterparts[0][e0]
            if size > 1:
                for e1 in range(len(counterparts[1])):
                    if counterparts[1][e1].individual == target[0].individual:
                        continue
                    target[1] = counterparts[1][e1]
                    if not self.assessTargetSet(target):
                        continue
                    if size > 2:
                        for e2 in range(len(counterparts[2])):
                                """
                            if isinstance(self.checkTree, list):
                                if self.checkTree[e0][e1][e2]:
                                    self.checkTree[e0][e1][e2] = False
                                    self.checkTree[e0][e2][e1] = False
                                    self.checkTree[e1][e0][e2] = False
                                    self.checkTree[e1][e2][e0] = False
                                    self.checkTree[e2][e0][e1] = False
                                    self.checkTree[e2][e1][e0] = False
                                    target[2] = counterparts[2][e2]
                                    if (counterparts[2][e2].individual == target[1].individual or 
                                        counterparts[2][e2].individual == target[0].individual):
                                        continue
                                    self.detectTransforms(source, target)
                            else:
                                """
                                target[2] = counterparts[2][e2]
                                self.detectTransforms(source, target)
                    else:
                            """
                        if isinstance(self.checkTree, list):
                            if self.checkTree[e0][e1]:
                                self.checkTree[e0][e1] = False
                                self.checkTree[e1][e0] = False
                                self.detectTransforms(source, target)
                        else:
                            """
                            self.detectTransforms(source, target)
        return self.transforms
    
    
    def detectTransforms(self, source, target):
        size = len(source)
        last = target[-1].individual
        if size > 1:
            first = target[0].individual
            if size > 2:
                second = target[1].individual
        toSet = [None, None, None]
            #########################################################
        
        if self.casus == similarityMap3D.casus.CASE_1PT_1LN_NOT_COLINEAR:
            toSet[1] = first.value
            u = last.direction
            v = toSet[1].subtract(last.root)

            if u.parallel(v):
                return                
            toSet[0] = last.root.add(u.scale(v))
            
            if toSet[0].subtract(toSet[1]).squareLength() == 0:
                return
            
            for i in target[-1].forms[target[-1].individual.sort.name]:
                check1 = False
                if (i.containsPoint(toSet[0]) >= self.subCase[0] and
                    (not(i.start == toSet[0] or i.end == toSet[0])) >= self.subCase[1]):
                    check1 = True
                    break
            if check1:
                d = toSet[1].subtract(toSet[0])
                
                toSet[2] = toSet[0].add(u.scale((d.dotproduct(d)/u.dotproduct(u))**c0p5*c2))
                
                for i in target[-1].forms[target[-1].individual.sort.name]:
                    if i.containsPoint(toSet[2]) >= self.subCase[2]:
                        tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection)
                        for tf in tfs:
                            if tf is not None:
                                self.transforms.append(tf)
                        break
    
                        
                toSet[2] = toSet[0].subtract(u.scale((d.dotproduct(d)/u.dotproduct(u))**c0p5*c2))
                for i in target[-1].forms[target[-1].individual.sort.name]:
                    if i.containsPoint(toSet[2]) >= self.subCase[2]:
                        tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection)
                        for tf in tfs:
                            if tf is not None:
                                self.transforms.append(tf)
                        break

        elif self.casus == similarityMap3D.casus.CASE_3PT_NOT_COLINEAR:
            if last == first or last == second or first == second:
                return
            if last.value.colinear(first.value, second.value):
                return
            toSet[0] = first.value
            toSet[1] = second.value
            toSet[2] = last.value

            tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection, True)
            for tf in tfs:
                if tf is not None:
                    self.transforms.append(tf)
            

        elif self.casus == similarityMap3D.casus.CASE_1PT_1ARC_NOT_COINCIDENT:
            if last.__class__ != self.subCase[0]:
                return
            lastCenter = last.position(last.center)
            if first.value == lastCenter:
                return
                
            toSet[0] = first.value
            toSet[1] = lastCenter

            u = last.normal
            v = first.value.subtract(lastCenter)
            
            if u.parallel(v):
                return
            
            cosAngle1 = u.unit().dotproduct(v.unit())    
            check1 = check2 = False
            if cosAngle1 == self.cosAngle:
                check1 = True
                if cosAngle1 == 0:
                    u1 = u.scale(-1)
                    check2 = True
            else:
                u1 = u.scale(-1)
                cosAngle2 = u1.unit().dotproduct(v.unit())
                if cosAngle2 == self.cosAngle:
                    check2 = True
                    
            if cosAngle1 == 0:
                for i in target[-1].forms[target[-1].individual.sort.name]:
                    containsCheck = False
                    if i.containsPoint(toSet[0]) >= self.subCase[1] :
                        containsCheck = True
                        break
                if not containsCheck:
                    return
                                
            if check1:
                toSet[2] = toSet[1].add(u.scale((v.dotproduct(v)/u.dotproduct(u))**c0p5*c2))
                tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection)
                for tf in tfs:
                    if tf is not None:
                        self.transforms.append(tf)

            if check2:
                toSet[2] = toSet[1].add(u1.scale((v.dotproduct(v)/u1.dotproduct(u1))**c0p5*c2))
                tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection)
                for tf in tfs:
                    if tf is not None:
                        self.transforms.append(tf)
                        
        elif self.casus == similarityMap3D.casus.CASE_2ARC_NOT_CONCENTRIC:
            if first.__class__ != self.subCase[0]:
                return
            if last.__class__ != self.subCase[1]:
                return
            firstCenter = first.position(first.center)
            lastCenter = last.position(last.center)
            
            toSet[0] = firstCenter
            toSet[1] = lastCenter
            u = last.normal
            v = firstCenter.subtract(lastCenter)
            
            if u.parallel(v):
                return
            
            toSet[2] = self.fromSet[1].add(u.scale((v.dotproduct(v)/u.dotproduct(u))**c0p5*c2))
            
            if first.__class__ == circle3D:
                firstRadius = first.radius
            else:
                firstRadius = first.radius[0]
            
            if last.__class__ == circle3D:
                lastRadius = last.radius
            else:
                lastRadius = last.radius[0]
            
            if not (firstRadius / lastRadius) == self.ratio[0]:
                return
                
            distanceVec = toSet[0].subtract(toSet[1])
            distanceSqr = distanceVec.dotproduct(distanceVec)
            
            if not ((firstRadius**2) / distanceSqr) == self.ratio[1]:
                return
                
            cosAngle1 = u.unit().dotproduct(v.unit())
            
            if cosAngle1 == self.cosAngle:
                check1 = True
                if cosAngle1 == 0:
                    u1 = u.scale(-1)
                    check2 = True
            else:
                u1 = u.scale(-1)
                cosAngle2 = u1.unit().dotproduct(v.unit())
                if cosAngle2 == self.cosAngle:
                    check2 = True
                                
            if check1:
                toSet[2] = toSet[1].add(u.scale((v.dotproduct(v)/u.dotproduct(u))**c0p5*c2))
                tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection)
                for tf in tfs:
                    if tf is not None:
                        self.transforms.append(tf)

            if check2:
                toSet[2] = toSet[1].add(u1.scale((v.dotproduct(v)/u1.dotproduct(u1))**c0p5*c2))
                tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection)
                for tf in tfs:
                    if tf is not None:
                        self.transforms.append(tf)
                        
        elif self.casus == similarityMap3D.casus.CASE_2ARC_NOT_PARALLEL:
            if first.__class__ != self.subCase[0]:
                return
            if last.__class__ != self.subCase[1]:
                return            
            firstCenter = first.position(first.center)

            if not firstCenter == last.position(last.center):
                return
            if first.normal.parallel(last.normal):
                return
            toSet[0] = firstCenter
            
            if first.__class__ == circle3D:
                firstRadius = first.radius
            else:
                firstRadius = first.radius[0]
            
            if last.__class__ == circle3D:
                lastRadius = last.radius
            else:
                lastRadius = last.radius[0]
            
            if not (firstRadius / lastRadius) == self.ratio :
                if (lastRadius / firstRadius) == self.ratio:
                    first, last = last, first
                    firstRadius, lastRadius = lastRadius, firstRadius
                else:
                    return

            firstUnit = first.normal.unit()
            lastUnit = last.normal.unit()
            revFirstUnit = firstUnit.scale(-1)
            revLastUnit = lastUnit.scale(-1)
            
            cosAngle1 = firstUnit.dotproduct(lastUnit)
            
            toSet[1] = firstUnit.scale(firstRadius)
            toSet[2] = lastUnit.scale(lastRadius)

            check1 = False
            check2 = False
            if cosAngle1 == self.cosAngle:
                check1 = True
                if cosAngle1 == 0:
                    check2 = True
            else:
                cosAngle2 = revFirstUnit.dotproduct(lastUnit)
                if cosAngle2 == self.cosAngle:
                    check2 = True
                    
            if check1:
                toSet[1] = toSet[0].add(firstUnit.scale(firstRadius))
                toSet[2] = toSet[0].add(lastUnit.scale(lastRadius))
                tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection, True)
                for tf in tfs:
                    if tf is not None:
                        self.transforms.append(tf)
                        
                toSet[1] = toSet[0].add(revFirstUnit.scale(firstRadius))
                toSet[2] = toSet[0].add(revLastUnit.scale(lastRadius))
                tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection, True)
                for tf in tfs:
                    if tf is not None:
                        self.transforms.append(tf)

            if check2:
                toSet[1] = toSet[0].add(revFirstUnit.scale(firstRadius))
                toSet[2] = toSet[0].add(lastUnit.scale(lastRadius))
                tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection, True)
                for tf in tfs:
                    if tf is not None:
                        self.transforms.append(tf)
                        
                toSet[1] = toSet[0].add(firstUnit.scale(firstRadius))
                toSet[2] = toSet[0].add(revLastUnit.scale(lastRadius))
                tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection, True)
                for tf in tfs:
                    if tf is not None:
                        self.transforms.append(tf)
                        
        if self.casus == similarityMap3D.casus.CASE_1ARC_1LN_NOT_COLINEAR:
            toSet[1] = first.position(first.center)
            u = last.direction
            v = toSet[1].subtract(last.root)

            if u.parallel(v):
                return                
            toSet[0] = last.root.add(u.scale(v))
            
            if toSet[0].subtract(toSet[1]).squareLength() == 0:
                return
            
            for i in target[-1].forms[target[-1].individual.sort.name]:
                check1 = False
                if (i.containsPoint(toSet[0]) >= self.subCase[0] and
                    (not(i.start == toSet[0] or i.end == toSet[0])) >= self.subCase[1]):
                    check1 = True
                    break
            if check1:
                d = toSet[1].subtract(toSet[0])
                
                toSet[2] = toSet[0].add(u.scale((d.dotproduct(d)/u.dotproduct(u))**c0p5*c2))
                
                for i in target[-1].forms[target[-1].individual.sort.name]:
                    if i.containsPoint(toSet[2]) >= self.subCase[2]:
                        tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection)
                        for tf in tfs:
                            if tf is not None:
                                self.transforms.append(tf)
                        break
    
                        
                toSet[2] = toSet[0].subtract(u.scale((d.dotproduct(d)/u.dotproduct(u))**c0p5*c2))
                for i in target[-1].forms[target[-1].individual.sort.name]:
                    if i.containsPoint(toSet[2]) >= self.subCase[2]:
                        tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection)
                        for tf in tfs:
                            if tf is not None:
                                self.transforms.append(tf)
                        break
                                
        elif self.casus == similarityMap3D.casus.CASE_3LN_NO_TWO_PARALLEL:
            if (last.parallel(first) or second.parallel(first) or second.parallel(last)):
                return
            try:
                is1 = first.intersection(last)
                is2 = first.intersection(second)
                is3 = second.intersection(last)
            except:
                return
            
            if is1 == is2 or is1 == is3 or is2 == is3:
                return
                
            u = is2.subtract(is1).squareLength()
            v = is3.subtract(is2).squareLength()
            w = is3.subtract(is1).squareLength()
            """
            u0= self.ratio[0]
            v0= self.ratio[1]
            w0= self.ratio[2]
            ts = []
            if (u/u0 == v/v0 and u/u0 == w/w0):
                ts.append([is1,is2,is3,target[0],target[1],target[2]])
            if (u/u0 == w/v0 and u/u0 == v/w0):
                ts.append([is2,is1,is3,target[0],target[2],target[1]])
            if (v/u0 == u/v0 and v/u0 == w/w0):
                ts.append([is3,is2,is1,target[1],target[0],target[2]])
            if (v/u0 == w/v0 and v/u0 == u/w0):
                ts.append([is2,is3,is1,target[1],target[2],target[0]])
            if (w/u0 == u/v0 and w/u0 == v/w0):
                ts.append([is3,is1,is2,target[2],target[0],target[1]])
            if (w/u0 == v/v0 and w/u0 == u/w0):
                ts.append([is1,is3,is2,target[2],target[1],target[0]])
            for toSet in ts:
            """
            if (u/self.ratio[0] == v/self.ratio[1] and u/self.ratio[0] == w/self.ratio[2]):
                toSet = [is1,is2,is3,target[0],target[1],target[2]]
                for i in toSet[3].forms[toSet[3].individual.sort.name]:
                    if (i.containsPoint(toSet[0]) >= self.subCase[0] and
                        i.containsPoint(toSet[1]) >= self.subCase[1]):
                            check = True
                            break
                    check = False
                if not check:
                    return
                for i in toSet[4].forms[toSet[3].individual.sort.name]:
                    if (i.containsPoint(toSet[1]) >= self.subCase[2] and
                        i.containsPoint(toSet[2]) >= self.subCase[3]):
                            check = True
                            break
                    check = False
                if not check:
                    return
                for i in toSet[5].forms[toSet[3].individual.sort.name]:
                    if (i.containsPoint(toSet[0]) >= self.subCase[4] and
                        i.containsPoint(toSet[2]) >= self.subCase[5]):
                            check = True
                            break
                    check = False
                if not check:
                    return
                tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection)
                for tf in tfs:
                    if tf is not None:
                        self.transforms.append(tf)
                        
            #########################################################

        elif self.casus == similarityMap3D.casus.CASE_3LN_FIRST_TWO_PARALLEL:
            if (last.parallel(first) or second.parallel(last)) or not second.parallel(first):
                return
            if second == first:
                return
                
            u = first.unitDir
            v = last.unitDir
            cosA = abs(u.dotproduct(v))
            if not self.cosAngle ==  cosA:
                return
            
            try:
                toSet[0] = first.intersection(last)
                toSet[1] = second.intersection(last)
            except ValueError:
                return
            
            ###########
            #------------------------------------------------------------    
            for i in target[0].forms[target[0].individual.sort.name]:
                if i.containsPoint(toSet[0]) >= self.subCase[0]:
                    check1 = True
                    break
                check1 = False
            for i in target[1].forms[target[1].individual.sort.name]:
                if i.containsPoint(toSet[1]) >= self.subCase[1]:
                    check2 = True
                    break
                check2 = False
            for i in target[2].forms[target[2].individual.sort.name]:
                if (i.containsPoint(toSet[0]) >= self.subCase[2] and
                    i.containsPoint(toSet[1]) >= self.subCase[3]):
                    check3 = True
                    break
                check3 = False

            if check1 and check2 and check3:        
                
                u = first.direction
                d = toSet[1].subtract(toSet[0])
                v = u.scale((d.dotproduct(d)/(u.dotproduct(u)))**c0p5*c2)
                toSet[2] = toSet[0].add(v)
                
                for i in target[0].forms[target[0].individual.sort.name]:
                    if i.containsPoint(toSet[2]) >= self.subCase[4]:                
                        tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection)
                        for tf in tfs:
                            if tf is not None:
                                self.transforms.append(tf) 
                        break
                    
                toSet[2] = toSet[0].subtract(v)
                for i in target[0].forms[target[0].individual.sort.name]:
                    if i.containsPoint(toSet[2]) >= self.subCase[4]:                
                        tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection)
                        for tf in tfs:
                            if tf is not None:
                                self.transforms.append(tf)
                        break
            #------------------------------------------------------------    
            """
            for i in target[0].forms[target[0].individual.sort.name]:
                if i.containsPoint(toSet[0]) >= self.subCase[1]:
                    check1 = True
                    break
                check1 = False
            for i in target[1].forms[target[1].individual.sort.name]:
                if i.containsPoint(toSet[1]) >= self.subCase[0]:
                    check2 = True
                    break
                check2 = False
            for i in target[2].forms[target[2].individual.sort.name]:
                if (i.containsPoint(toSet[0]) >= self.subCase[3] and
                    i.containsPoint(toSet[1]) >= self.subCase[2]):
                    check3 = True
                    break
                check3 = False
            
            if check1 and check2 and check3:  
                
                u = second.direction
                d = toSet[0].subtract(toSet[1])
                v = u.scale((d.dotproduct(d)/(u.dotproduct(u)))**c0p5*c2)
                toSet[2] = toSet[1].add(v)
                                
                for i in target[1].forms[target[1].individual.sort.name]:
                    if i.containsPoint(toSet[2]) >= self.subCase[4]:                
                        tfs = affineTransform3D.similarityTransform(self.fromSet, [toSet[1],toSet[0], toSet[2]], self.reflection)
                        for tf in tfs:
                            if tf is not None:
                                self.transforms.append(tf)
                
                toSet[2] = toSet[1].subtract(v)
                
                for i in target[1].forms[target[1].individual.sort.name]:
                    if i.containsPoint(toSet[2]) >= self.subCase[4]:                
                        tfs = affineTransform3D.similarityTransform(self.fromSet, [toSet[1],toSet[0], toSet[2]], self.reflection)
                        for tf in tfs:
                            if tf is not None:
                                self.transforms.append(tf)
            """
            #------------------------------------------------------------    
            #########################################################
            #########################################################
            #########################################################
            #########################################################
            #########################################################
        elif self.casus == similarityMap3D.casus.CASE_2LN_SKEWED:
            if first.parallel(last):
                return
            elif not first.skew(last):
                return
            n = last.direction.product(first.direction)
            p1 = first.root
            p2 = last.root
            d1 = first.direction
            d2 = last.direction
            n1 = d1.product(n)
            n2 = d2.product(n)
            toSet[0] = p1.add(d1.scale((p2.subtract(p1).dotproduct(n2))/(d1.dotproduct(n2))))
            toSet[1] = p2.add(d2.scale((p1.subtract(p2).dotproduct(n1))/(d2.dotproduct(n1))))
            
            check1 = False
            for i in target[0].forms[target[0].individual.sort.name]:
                check = i.containsPoint(toSet[0])
                if check >= self.subCase[0]:
                    check1 = True
                    
            check2 = False
            for i in target[1].forms[target[1].individual.sort.name]:
                check = i.containsPoint(toSet[1])
                if check >= self.subCase[1]:
                    check2 = True
                       

            if not (check1 and check2):
                return
            
            d = toSet[1].subtract(toSet[0])
            
            allToSet = []
            
            if check1 and check2:
                v1 = d1.scale((d.dotproduct(d)/(d1.dotproduct(d1)))**c0p5*c2)
                toSet2_1 = toSet[0].add(v1)
                toSet2_2 = toSet[0].subtract(v1)
                for i in target[0].forms[target[0].individual.sort.name]:
                    if i.containsPoint(toSet2_1) >= self.subCase[2]:
                        allToSet.append([toSet[0], toSet[1], toSet2_1])
                        break
                for i in target[0].forms[target[0].individual.sort.name]:
                    if i.containsPoint(toSet2_2) >= self.subCase[2]:
                        allToSet.append([toSet[0], toSet[1], toSet2_2])
                        break
                    
            for i in allToSet:
                tfs = affineTransform3D.similarityTransform(self.fromSet, i, self.reflection)
                for tf in tfs:
                    if tf is not None:
                        self.transforms.append(tf)


            #########################################################
            #########################################################
            #########################################################
            #########################################################
            #########################################################
            #########################################################
            
        elif self.casus == similarityMap3D.casus.CASE_2PT_1PL_NOT_PERPENDICULAR:
            foot = last.projection(first.value)
            if foot == first: 
                if self.subCase == 1:
                    return
                else:
                    foot = last.projection(second.value)
                    if foot == second:
                        if self.subCase == 2:
                            return
                        else:
                            d = second.subtract(first)
                            u = last.normal.product(d)
                            foot = first.add(u.scale((d.dotProduct(d) / (u.dotProduct(u)))**c0p5*c2))

            if (foot.colinear(second.value, first.value)):
                return
                
            toSet[0] = first.value
            toSet[1] = second.value
            toSet[2] = foot
            
            tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection)
            for tf in tfs:
                if tf is not None:
                    self.transforms.append(tf)


        elif self.casus == similarityMap3D.casus.CASE_2LN_1PL_PARALLEL_BOTH:
            if not (last.parallelLine(first) and last.parallelLine(second)):
                return
                
            toSet[0] = first.intersection(second)
            toSet[1] = last.projection(toSet[0])
            d0 = last.normal.product(first.unitDir)
            l0 = toSet[1].subtract(toSet[0]).length()
            
            toSet[2] = toSet[1].add(d0.scale(l0))
            tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection)
            for tf in tfs:
                if tf is not None:
                    self.transforms.append(tf)
                    
            toSet[2] = toSet[1].subtract(d0.scale(l0))
            tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection)
            for tf in tfs:
                if tf is not None:
                    self.transforms.append(tf)
        elif self.casus == similarityMap3D.casus.CASE_2LN_1PL_PARALLEL_FIRST:
            if (not last.parallelLine(first)) or last.parallelLine(second) or first.skew(second):
                return
            toSet[0] = last.intersectionLine(second)
            toSet[1] = first.intersection(second)
            l0 = toSet[1].subtract(toSet[0]).length()
            
            toSet[2] = toSet[1].add(first.unitDir.scale(l0))
            tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection)
            for tf in tfs:
                if tf is not None:
                    self.transforms.append(tf)
                    
            toSet[2] = toSet[1].subtract(first.unitDir.scale(l0))
            tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection)
            for tf in tfs:
                if tf is not None:
                    self.transforms.append(tf)
        elif self.casus == similarityMap3D.casus.CASE_2LN_1PL_NOT_PARALLEL_NOR_CONCURRENT:
            if last.parallelLine(first) or last.parallelLine(second):
                return
            toSet[0] = last.intersectionLine(first)
            toSet[1] = last.intersectionLine(second)
            
            if toSet[0] == toSet[1]:
                return
                
            toSet[2] = toSet[0].add(last.normal.scale(toSet[0].subtract(toSet[1]).length()))
            tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection)
            for tf in tfs:
                if tf is not None:
                    self.transforms.append(tf)
                    
            toSet[2] = toSet[0].subtract(last.normal.scale(toSet[0].subtract(toSet[1]).length()))
            tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection)
            for tf in tfs:
                if tf is not None:
                    self.transforms.append(tf)
                    
        elif self.casus == similarityMap3D.casus.CASE_1LN_2PL_NOT_ALL_CONCURRENT:
            if last.parallelLine(first) != self.subCase:
                return

            toSet[0] = second.intersectionLine(first)
            toSet[1] = last.projection(toSet[0])
                
            if toSet[0] == toSet[1]:
                return similarityMap3D.distinguishes.UNSUITABLE
            l0 = toSet[1].subtract(toSet[0]).length()
            
            toSet[2] = toSet[0].add(first.unitDir.scale(l0))
            tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection)
            for tf in tfs:
                if tf is not None:
                    self.transforms.append(tf)
                    
            toSet[2] = toSet[0].subtract(first.unitDir.scale(l0))
            tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection)
            for tf in tfs:
                if tf is not None:
                    self.transforms.append(tf)
        elif self.casus == similarityMap3D.casus.CASE_1PT_2PL_NOT_CONCURRENT:
            if last.parallel(second):
                return similarityMap3D.distinguishes.UNSUITABLE
            toSet[0] = first.value
            toSet[1] = second.projection(first.value)
            toSet[2] = last.projection(first.value)
            tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection)
            for tf in tfs:
                if tf is not None:
                    self.transforms.append(tf)
        return
        
        
    def specialPotentialTransforms(self):
        if len(self.distinguishables) == 1:
            dist = self.distinguishables[0]
            if dist.element.individual.type == 'ind-point3D':
                fromVect = dist.element.individual.value
                for i in dist.counterparts:
                    toVect = i.individual.value
                    self.transforms.append(affineTransform3D.translation(toVect.subtract(fromVect)))
                    
            if dist.element.individual.type == 'ind-line3D':
                fromVects = [dist.element.individual.assoc.start, dist.element.individual.assoc.end]
                dirFrom = fromVects[1].subtract(fromVects[0])
                fromSqLen = dirFrom.squareLength()
                for i in dist.counterparts:
                    toVects = [i.individual.assoc.start, i.individual.assoc.end]

                    dirTo = toVects[1].subtract(toVects[0])
                    toSqLen = dirTo.squareLength()
                    axis = dirFrom.product(dirTo)
                    
                    
                    sqLengths = fromSqLen * toSqLen
                    scalarSq = toSqLen/ fromSqLen
                    cosSq = dirFrom.dotproduct(dirTo) ** c2 / sqLengths
                    
                    sinSq = axis.dotproduct(axis) / sqLengths
                    
                    if axis.isZero():
                        baseRotation = affineTransform3D.scale(scalarSq.sqrt())
                        baseRotation1 = baseRotation.tScale(-1)
                    else: 
                        baseRotation = affineTransform3D.scaledRotation(axis, sinSq, cosSq, dirFrom.dotproduct(dirTo).sign(), scalarSq)
                        dirTo = dirTo.inverse()
                        axis = axis.inverse()
                        baseRotation1 = affineTransform3D.scaledRotation(axis, sinSq, cosSq, dirFrom.dotproduct(dirTo).sign(), scalarSq)
                    transformation = affineTransform3D.translation(toVects[0].subtract(fromVects[0].transform(baseRotation))).combine(baseRotation)
                    transformation1 = affineTransform3D.translation(toVects[1].subtract(fromVects[0].transform(baseRotation1))).combine(baseRotation1)
                    self.transforms.append(transformation)
                    self.transforms.append(transformation1)
        if len(self.distinguishables) == 2:
            dist0 = self.distinguishables[0]
            dist1 = self.distinguishables[1]
            if dist0.element.individual.type == dist1.element.individual.type == 'ind-point3D':
                fromVects = [dist0.element.individual.value, dist1.element.individual.value]
                
                dirFrom = fromVects[1].subtract(fromVects[0])
                fromSqLen = dirFrom.squareLength()
                for i in dist0.counterparts:
                    for j in dist1.counterparts:
                        if i == j:
                            continue
                        toVects = [i.individual.value, j.individual.value]
    
                        dirTo = toVects[1].subtract(toVects[0])
                        toSqLen = dirTo.squareLength()
                        axis = dirFrom.product(dirTo)
                        
                        
                        sqLengths = fromSqLen * toSqLen
                        scalarSq = toSqLen/ fromSqLen
                        cosSq = dirFrom.dotproduct(dirTo) ** c2 / sqLengths
                        
                        sinSq = axis.dotproduct(axis) / sqLengths
                        
                        if axis.isZero():
                            baseRotation = affineTransform3D.scale(scalarSq.sqrt())
                            baseRotation1 = baseRotation.tScale(-1)
                        else: 
                            baseRotation = affineTransform3D.scaledRotation(axis, sinSq, cosSq, dirFrom.dotproduct(dirTo).sign(), scalarSq)
                            dirTo = dirTo.inverse()
                            axis = axis.inverse()
                            baseRotation1 = affineTransform3D.scaledRotation(axis, sinSq, cosSq, dirFrom.dotproduct(dirTo).sign(), scalarSq)
                        transformation = affineTransform3D.translation(toVects[0].subtract(fromVects[0].transform(baseRotation))).combine(baseRotation)
                        transformation1 = affineTransform3D.translation(toVects[1].subtract(fromVects[0].transform(baseRotation1))).combine(baseRotation1)
                        self.transforms.append(transformation)
                        self.transforms.append(transformation1)
            elif dist0.element.individual.type == dist1.element.individual.type == 'ind-line3D':
                if dist1.element.individual.parallel(dist0.element.individual):
                    fromVects = []
                    fromVects.append(dist0.element.forms[dist0.element.individual.assoc.sort.name][0].start)
                    fromVects.append(dist1.element.forms[dist1.element.individual.assoc.sort.name][0].start)
                    fromVects.append(dist1.element.forms[dist1.element.individual.assoc.sort.name][0].end)
                    for i in dist0.counterparts:
                        for j in dist1.counterparts:
                            if i.individual == j.individual:
                                continue
                            if not i.individual.parallel(j.individual):
                                continue
                            for line0 in i.forms[dist0.element.individual.assoc.sort.name]:
                                for line1 in j.forms[dist1.element.individual.assoc.sort.name]:
                                    toVects = [line0.start,line1.start,line1.end]
                                    tfs = affineTransform3D.similarityTransform(fromVects, toVects, 1)
                                    for tf in tfs:
                                        if tf is not None:
                                            self.transforms.append(tf)
                                    toVects = [line0.end,line1.end,line1.start]
                                    tfs = affineTransform3D.similarityTransform(fromVects, toVects, 1)
                                    for tf in tfs:
                                        if tf is not None:
                                            self.transforms.append(tf)
                                            
                else:
                    fromVects = [dist0.element.individual.intersection(dist1.element.individual)]
                    checkContains = False
                    for i in dist0.element.forms[dist0.element.individual.assoc.sort.name]:
                        if i.containsPoint(fromVects[0]):
                            checkContains = True
                            pt = i.start if not i.start == fromVects[0] else i.end
                            fromVects.append(pt)
                    if not checkContains:
                        l = dist0.element.forms[dist0.element.individual.assoc.sort.name][0]
                        pt = l.start
                        fromVects.append(pt)
                    checkContains = False
                    for i in dist1.element.forms[dist1.element.individual.assoc.sort.name]:
                        if i.containsPoint(fromVects[0]):
                            checkContains = True
                            pt = i.start if not i.start == fromVects[0] else i.end
                            fromVects.append(pt)
                    if not checkContains:
                        l = dist1.element.forms[dist1.element.individual.assoc.sort.name][0]
                        pt = l.start
                        fromVects.append(pt)
                        
                    for i in dist0.counterparts:
                        for j in dist1.counterparts:
                            if not i.individual.intersect(j.individual):
                                continue
                            toVects = [i.individual.intersection(j.individual), None, None]
                            for line0 in i.forms[dist0.element.individual.assoc.sort.name]:
                                for line1 in j.forms[dist1.element.individual.assoc.sort.name]:
                                    if not toVects[0] == line0.start:
                                        toVects[1] = line0.start
                                        if not toVects[0] == line1.start:
                                            toVects[2] = line1.start
                                            tfs = affineTransform3D.similarityTransform(fromVects, toVects, 1)
                                            for tf in tfs:
                                                if tf is not None:
                                                    self.transforms.append(tf)
                                        if not toVects[0] == line1.end:
                                            toVects[2] = line1.end
                                            tfs = affineTransform3D.similarityTransform(fromVects, toVects, 1)
                                            for tf in tfs:
                                                if tf is not None:
                                                    self.transforms.append(tf)
                                    if not toVects[0] == line0.end:
                                        toVects[1] = line0.end
                                        if not toVects[0] == line1.start:
                                            toVects[2] = line1.start
                                            tfs = affineTransform3D.similarityTransform(fromVects, toVects, 1)
                                            for tf in tfs:
                                                if tf is not None:
                                                    self.transforms.append(tf)
                                        if not toVects[0] == line1.end:
                                            toVects[2] = line1.end
                                            tfs = affineTransform3D.similarityTransform(fromVects, toVects, 1)
                                            for tf in tfs:
                                                if tf is not None:
                                                    self.transforms.append(tf)
                
        return self.transforms
        