# -*- coding: utf-8 -*-
"""
Created on Fri Aug 12 14:33:25 2016

@author: akibdpt
"""
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import absolute_import
from builtins import str
"""
from future import standard_library
standard_library.install_aliases()
"""
from builtins import object
from .coordinate import coordinate, c1, c0p5, c2, acos, pi, atan2
from ..io.parseReader import parseReader, ParsingError

"""
 * A 2D vector specifies a position in a two-dimensional cartesian space.
 * If normalized, it only specifies a direction.
 * 
 * The 2D Vector class defines a vector as a pair of coordinates
 * and a w factor to reflect the vector's infinity characteristic
 * A 2D Vector object is never modified after creation, thus,
 * it can be used multiple times.
"""
class vector2D(object):
    """
    initialize vector
    containing x coordinate, y coordinate and w
    w is indicator of infinity. If w is 1 -> this is not an infinity vector
                                If w is 0 -> this is an infinity vector with direction(x,y)
    """
    def __init__(self, x, y, infinity = False):
        self.type = 'struct-vector2D'
        if infinity:
            self.w = 0
        else:
            self.w = 1
        self.x = coordinate(x)
        self.y = coordinate(y)

    """
    return the string that represents this vector's values
    """       
    def __str__(self):
        if self.w == 0:
            return ' inf('+str(self.x)+', '+str(self.y)+')'
        return '('+str(self.x)+', '+str(self.y)+')'
        
    def __repr__(self):
        return self.__str__()
        
    def __hash__(self):
        s = str(self.x)+' '+str(self.y)
        return hash(s)
        
    def __abs__(self):
        if self.sign() >= 0:
            return self
        else:
            return vector2D(-self.x, -self.y)
        
    def printVal(self, displacement = 0):
        if self.w == 0:
            return 'inf('+self.x+','+self.y+')'
        return '('+str(self.x+displacement)+','+str(self.y)+')'
        
    def printRhino(self, string = '', displacement = 0):
        return '_Point '+str(self.x + displacement)+','+str(self.y)
    
    """
    find the sign of the vector
    the vector is positive if the x value is positive
    the vector is negative if the x value is negative
    if x value is 0, the sign of the vector is decided by the y value
    return 1 if positive, -1 if negative and 0 if both x and y are 0
    """
    def sign(self):
        if self.x > 0:
            return 1
        if self.x < 0:
            return -1
        if self.y > 0:
            return 1
        if self.y < 0:
            return -1
        return 0
    
    """
    check if this vector is a zero vector: (0,0)
    """
    def isZero(self):
        return self.x == 0 and self.y == 0
        
    """
    checking if two vectors are parallel
    two vectors are parallel if: this.x / other.x = this.y / other.y
                             --> this.x * other.y = this.y * other.x
    """
    def parallel(self,other):
        if other.type != 'struct-vector2D':
            return False
        return (self.x * other.y) == (self.y * other.x)
        
    """
    checking if two vectors are perpendicular with one another
    two vectors are perpendicular if their dot product value is 0
    """
    def perpendicular(self,other):
        if other.type != 'struct-vector2D':
            return False
        return self.dotproduct(other) == 0
        
    def isInfinite(self):
        return self.w == 0
        
    """
    checking if this vector is equal to other vector
    two vectors are equal if both have equal x values, equal y values and equal z values
    """
    def __eq__(self, other):
        if other is None or other.__class__ != self.__class__:
            return False
        return self.x == other.x  and self.y  == other.y and self.w == other.w
        
    def __ne__(self, other):
        return not self == other
 
        
    def __lt__(self, other):
        return self.compareValue(other) == -1
        
    def __le__(self, other):
        return self.compareValue(other) < 1
        
    def __gt__(self, other):
        return self.compareValue(other) == 1
        
    def __ge__(self, other):
        return self.compareValue(other) > -1

    """
    comparing the value of this vector with another vector
    a vector is greater if its x value is greater and vice versa
    if the x values of the two vectors are the same, the vector with the 
    greater y value is greater
    
    return 1 if this vector is greater than the other vector
    return -1 if this vector is lesser than the other vector
    return 0 if both vectors have equal x and y values
    """
    def compareValue(self, other):
        if self.w == other.w:
            if self.x > other.x:
                return 1
            elif self.x < other.x:
                return -1
            elif self.y > other.y:
                return 1
            elif self.y < other.y:
                return -1
            return 0
        elif self.w == 0:
            if self.x < 0:
                return -1
            elif self.x > 0:
                return 1
            elif other.x < 0:
                return 1
            elif other.x > 0:
                return -1
            else:
                if self.y < 0:
                    return -1
                elif self.y > 0:
                    return 1
                elif other.y < 0:
                    return 1
                elif other.y > 0:
                    return -1
                else:
                    return 0
        elif other.w == 0:
            if other.x < 0:
                return 1
            elif other.x > 0:
                return -1
            elif self.x < 0:
                return -1
            elif self.x > 0:
                return 1
            else:
                if other.y < 0:
                    return 1
                elif other.y > 0:
                    return -1
                elif self.y < 0:
                    return -1
                elif self.y > 0:
                    return 1
                else:
                    return 0
    
    """
    scaling the vector by a certain factor
    if factor is a number: multiply the vector's x and y with the factor
    if factor is another vector: multiply the vector's x and y with the projection length of the other vector
    """    
    def scale(self, other):
        if isinstance(other, coordinate) or isinstance(other, int):
            return vector2D(self.x*other, self.y*other)
        else:
            scalar = (other.dotproduct(self)) / (self.dotproduct(self))
            return vector2D(self.x*scalar, self.y*scalar)
            
    """
    find the vector with absolute values of this vector's x and y
    """
    def abs(self):
        return self.scale(self.sign())
                
    """
    find the vector that is a reflection of this vector by the origin
    """
    def inverse(self):
        return vector2D(-self.x, -self.y)
        
    """
    find the length of the vector
    """
    def length(self):
        return (self.x * self.x + self.y *self.y) ** c0p5
        
    """
    find the squared length of the vector
    *if length equality is to be checked, use squared length since it is more 
     time efficient (no square root calculation)
    """
    def squareLength(self):
        return (self.x * self.x + self.y *self.y)
        
    """
    find the unit vector of the vector: dividing the vector's x and y by the length of the vector
    """
    def unit(self):
        a = c1 / self.length()
        return vector2D(self.x * a, self.y * a)
    
    """
    adding two vector
    """
    def add(self,other):
        if self.w == 1 and other.w == 1:
            return vector2D(self.x+other.x, self.y+other.y)
        if self.w == 0:
            if other.w == 0:
                raise ValueError('Adding two infinite vectors is undefined')
            else:
                return self
        return other
    
    """
    subtracting the vector by another vector
    """
    def subtract(self, other):
        if self.w == 1 and other.w == 1:
            return vector2D(self.x-other.x, self.y-other.y)
        elif self.w == 0:
            if other.w == 0 and self != other.inverse():
                return vector2D(self.x-other.x, self.y-other.y, True)
            else:
                return self
        return other
        
    """
    find the dot product of the vector with another
    """
    def dotproduct(self, other):
        return self.x * other.x + self.y * other.y
        
    """
    find the cross product of the vector with another,
    since the vectors are in 2D space, the vectors will be (x0,y0,0) and (x1,y1,0)
    in 3D space --> the product vector will be (0, 0, x0*y1-x1*y0)
    this function returns only the z value of the product vector since the vector
    itself cannot be represented in the 2D space
    """
    def product(self, other):
        return self.x * other.y - self.y * other.x
    
    """    
    """
    def scalar(self, other):
        return other.dotproduct(self) / self.dotproduct(self)
    
    """
    check if this vector and 2 other vectors are colinear
    """
    def colinear(self, other1, other2):
        v1 = other1.subtract(self)
        v2 = other2.subtract(self)
        return (v1.x * v2.y) == (v1.y * v2.x)
        
    """
    find the midpoint vector between two vectors
    """
    def midpoint(self, other):
        return vector2D((self.x+other.x)/c2,(self.y+other.y)/c2)
    
    """
    find the clockwise angle from this vector to another vector
    """
    def angle(self, other):
        a = self.unit()
        b = other.unit()
        return (atan2(b.y.value, b.x.value) - atan2(a.y.value, a.x.value))
        c = atan2(b.y.value, b.x.value) - atan2(a.y.value, a.x.value)
        if (c < 0 and c > -pi) or (c > pi and c < 2*pi):            
            return - acos(self.unit().dotproduct(other.unit()).value)
        return acos(self.unit().dotproduct(other.unit()).value)
                
    """
    transforming the vector with a transformation matrix
    matrix: [[a, b, c]
             [d, e, f]]
    this vector: (x, y)
    => vector after transforming: (a*x + b*y + c, d*x + e*y + f)
    """
    def transform(self, transf):
        return vector2D((self.x*transf.matrix[0][0] + self.y*transf.matrix[0][1] + transf.matrix[0][2]),
                        (self.x*transf.matrix[1][0] + self.y*transf.matrix[1][1] + transf.matrix[1][2]))
                 
    """
    parsing vector from a string
    input: string e.g. (0,2)
    output: 2 coordinate values x and y e.g. 0 and 2
    """
    @staticmethod
    def readVector(reader):
        if reader.newToken() == parseReader.MINUS:
            neg = '-'
            reader.newToken()
        else:
            neg = ''
        x = coordinate(neg+reader.value())
        
        reader.newToken() # comma
        
        if reader.newToken() == parseReader.MINUS:
            neg = '-'
            reader.newToken()
        else:
            neg = ''
        y = coordinate(neg+reader.value())
        
        reader.newToken() # bracket
        
        reader.newToken() # next token
        return x,y
                                                

    
"""
 * A 3D vector specifies a position in a three-dimensional cartesian space.
 * If normalized, it only specifies a direction.
 * 
 * The 3D Vector class defines a vector as a set of three coordinates
 * and a w factor to reflect the vector's infinity characteristic
 * A 3D Vector object is never modified after creation, thus,
 * it can be used multiple times.
"""
class vector3D(object):
    
    """
    initialize vector
    containing x coordinate, y coordinate, z coordinate and w
    w is indicator of infinity. If w is 1 -> this is not an infinity vector
                                If w is 0 -> this is an infinity vector with direction(x,y)
    """
    def __init__(self, x, y, z=0, infinity = False):
        self.type = 'struct-vector3D'
        if infinity:
            self.w = 0
        else:
            self.w = 1
        self.x = coordinate(x)
        self.y = coordinate(y)
        self.z = coordinate(z)
            
    """
    return the string that represents this vector's values
    """       
    def __str__(self):
        if self.w == 0:
            return ' inf('+str(self.x)+', '+str(self.y)+', '+str(self.z)+')'
        return '('+str(self.x)+', '+str(self.y)+', '+str(self.z)+')'
        
    def __repr__(self):
        return self.__str__()
        
    def __hash__(self):
        s = str(self.x)+' '+str(self.y)+' '+str(self.z)
        return hash(s)
        
    def __abs__(self):
        if self.sign() >= 0:
            return self
        else:
            return vector3D(-self.x, -self.y, -self.z)
        
    def printVal(self, displacement = 0):
        if self.w == 0:
            return 'inf('+self.x+','+self.y+','+self.z+')'
        return '('+str(self.x+displacement)+','+self.y+','+self.z+')'
        
    def printRhino(self, string = '', displacement = 0):
        return '_Point '+str(self.x + displacement)+','+str(self.y)+','+str(self.z)
    

    """
    find the sign of the vector
    the vector is positive if the x value is positive
    the vector is negative if the x value is negative
    if x value is 0, the sign of the vector is decided by the y value
    if x value and y value is 0, the sign of the vector is decided by the z value
    return 1 if positive, -1 if negative and 0 if both x, y and z are 0
    """
    def sign(self):
        if self.x > 0:
            return 1
        if self.x < 0:
            return -1
        if self.y > 0:
            return 1
        if self.y < 0:
            return -1
        if self.z > 0:
            return 1
        if self.z < 0:
            return -1
        return 0
    
    """
    check if this vector is a zero vector: (0,0,0)
    """
    def isZero(self):
        return self.x.isZero() and self.y.isZero() and self.z.isZero()
        
    """
    checking if two vectors are parallel
    
    two vectors are parallel: 
        if this.x and other.x is 0:     this.y / other.y = this.z / other.z 
                                    --> this.y * other.z = this.z * other.y
                                    
        else:                           this.x / other.x = this.y / other.y = this.z / other.z
                                    --> this.x * other.y = this.y * other.x 
                                    and this.x * other.z = this.z * other.x
    """
    def parallel(self,other):
        if other.type != 'struct-vector3D':
            return False
        if self.x == 0 and other.x == 0:
            lhs = (self.z * other.y)
            rhs = (self.y * other.z)
            return lhs == rhs
            
        lhs1 = (self.x * other.y)
        rhs1 = (self.y * other.x)
        
        lhs2 = (self.x * other.z)
        rhs2 = (self.z * other.x)
        
        return lhs1 == rhs1 and lhs2 == rhs2
        
    """
    checking if two vectors are perpendicular with one another
    two vectors are perpendicular if their dot product value is 0
    """
    def perpendicular(self,other):
        if other.type != 'struct-vector3D':
            return False
        return self.dotproduct(other) == 0
        
    def isInfinite(self):
        return self.w == 0
        
    """
    checking if this vector is equal to other vector
    two vectors are equal if both have equal x values, equal y values, 
                                       equal z values and equal w values
    """
    def __eq__(self, other):
        if other is None or other.__class__ != self.__class__:
            return False
        return (self.x == other.x  and self.y  == other.y and 
                self.z  == other.z and self.w == other.w)
        
    def __ne__(self, other):
        return not self == other
 
    def __lt__(self, other):
        return self.compareValue(other) == -1
        
    def __le__(self, other):
        return self.compareValue(other) < 1

    def __gt__(self, other):
        return self.compareValue(other) == 1
        
    def __ge__(self, other):
        return self.compareValue(other) > -1

    """
    comparing the value of this vector with another vector
    a vector is greater if its x value is greater and vice versa
    if the x values of the two vectors are the same, the vector with the 
    greater y value is greater, if y values are the same, the vector with the 
    greater z value is greater
    
    return 1 if this vector is greater than the other vector
    return -1 if this vector is lesser than the other vector
    return 0 if both vectors have equal x and y values
    """
    def compareValue(self, other):
        if self.w == other.w:
            if self.x > other.x:
                return 1
            elif self.x < other.x:
                return -1
            elif self.y > other.y:
                return 1
            elif self.y < other.y:
                return -1
            elif self.z > other.z:
                return 1
            elif self.z < other.z:
                return -1
            return 0
        elif self.w == 0:
            if self.x < 0:
                return -1
            elif self.x > 0:
                return 1
            elif other.x < 0:
                return 1
            elif other.x > 0:
                return -1
            else:
                if self.y < 0:
                    return -1
                elif self.y > 0:
                    return 1
                elif other.y < 0:
                    return 1
                elif other.y > 0:
                    return -1
                else:
                    if self.z < 0:
                        return -1
                    elif self.z > 0:
                        return 1
                    elif other.z < 0:
                        return 1
                    elif other.z > 0:
                        return -1
                    else:
                        return 0
        elif other.w == 0:
            if other.x < 0:
                return 1
            elif other.x > 0:
                return -1
            elif self.x < 0:
                return -1
            elif self.x > 0:
                return 1
            else:
                if other.y < 0:
                    return 1
                elif other.y > 0:
                    return -1
                elif self.y < 0:
                    return -1
                elif self.y > 0:
                    return 1
                else:
                    if other.z < 0:
                        return 1
                    elif other.z > 0:
                        return -1
                    elif self.z < 0:
                        return -1
                    elif self.z > 0:
                        return 1
                    else:
                        return 0
    
    """
    scaling the vector by a certain factor
    if factor is a number: multiply the vector's x and y with the factor
    if factor is another vector: multiply the vector's x and y with the projection length of the other vector
    """    
    def scale(self, other):
        if isinstance(other, coordinate) or isinstance(other, int):
            return vector3D(self.x*other, self.y*other, self.z*other)
        else:
            scalar = (other.dotproduct(self)) / (self.dotproduct(self))
            return vector3D(self.x*scalar, self.y*scalar, self.z*scalar)
            
    """
    find the vector with absolute values of this vector's x and y
    """
    def abs(self):
        return self.scale(self.sign())
        
        
    """
    find the vector that is a reflection of this vector by the origin
    """
    def inverse(self):
        return vector3D(-self.x, -self.y, -self.z)
        
    """
    find the length of the vector
    """
    def length(self):
        return (self.x * self.x + self.y *self.y + self.z *self.z) ** c0p5
        
    """
    find the squared length of the vector
    *if length equality is to be checked, use squared length since it is more 
     time efficient (no need to calculate square root)
    """
    def squareLength(self):
        return (self.x * self.x + self.y *self.y + self.z *self.z)
        
    """
    find the unit vector of the vector: dividing the vector's x and y by the length of the vector
    """
    def unit(self):
        a = c1 / self.length()
        return vector3D(self.x * a, self.y * a, self.z * a)
    
    """
    adding two vector
    """
    def add(self,other):
        if self.w == 1 and other.w == 1:
            return vector3D(self.x+other.x, self.y+other.y, self.z+other.z)
        if self.w == 0:
            if other.w == 0:
                raise ValueError('Adding two infinite vectors is undefined')
            else:
                return self
        return other
    
    """
    subtracting the vector by another vector
    """
    def subtract(self, other):
        if self.w == 1 and other.w == 1:
            return vector3D(self.x-other.x, self.y-other.y, self.z-other.z)
        if self.w == 0:
            if other.w == 0 and self != other.inverse():
                return vector3D(self.x-other.x, self.y-other.y, self.z-other.z, True)
            else:
                return self
        return other
        
    """
    find the dot product of the vector with another
    """
    def dotproduct(self, other):
        return self.x * other.x + self.y * other.y + self.z * other.z
        
    """
    find the product of the vector with another,
    since the vectors are in 2D space, the vectors will be (x0,y0,0) and (x1,y1,0)
    in 3D space --> the product vector will be (0, 0, x0*y1-x1*y0)
    this function return only the z value of the product vector since the vector
    itself cannot be represented in the 2D space
    """
    def product(self, other):
        return vector3D(self.y * other.z - self.z * other.y,
                        self.z * other.x - self.x * other.z,
                        self.x * other.y - self.y * other.x)
                        
    """
    find the sign of the cross product of this vector with another vector
    """
    def productSign(self, other):
        c = self.y * other.z - self.z * other.y
        if c > 0:
            return 1
        elif c < 0:
            return -1
        c = self.z * other.x - self.x * other.z
        if c > 0:
            return 1
        elif c < 0:
            return -1
        c = self.x * other.y - self.y * other.x
        if c > 0:
            return 1
        elif c < 0:
            return -1
        return 0
    
    """    
    """
    def scalar(self, other):
        return other.dotproduct(self) / self.dotproduct(self)
    
    """
    check if this vector and 2 other vectors are colinear
    """
    def colinear(self, other1, other2):
        return other1.subtract(self).parallel(other2.subtract(self))
        
    """
    find the midpoint vector between two vectors
    """
    def midpoint(self, other):
        return vector3D((self.x+other.x)/c2, (self.y+other.y)/c2, (self.z+other.z)/c2)
        
    
    
    ###########################################################################
    """
    def angle(self, other):
        a = self.unit()
        b = other.unit()
        #return (atan2(b.y, b.x) - atan2(a.y, a.x))
        
        c = atan2(b.y, b.x) - atan2(a.y, a.x)
        if (c < 0 and c > -pi) or (c > pi and c < 2*pi):            
            return - acos(self.unit().dotproduct(other.unit()).value)
        return acos(self.unit().dotproduct(other.unit()).value)
    """
    ###########################################################################

    """
    transforming the vector with a transformation matrix
    matrix: [[a, b, c, d]
             [e, f, g, h]
             [i, j, k, l]]
    this vector: (x, y, z)
    => vector after transforming: (a*x + b*y + c*z +d, e*x + f*y + g*z +h, i*x + j*y + k*z +l)
    """
    def transform(self, transf):
        return vector3D((self.x*transf.matrix[0][0] + self.y*transf.matrix[0][1] + self.z*transf.matrix[0][2] + transf.matrix[0][3]),
                        (self.x*transf.matrix[1][0] + self.y*transf.matrix[1][1] + self.z*transf.matrix[1][2] + transf.matrix[1][3]),
                        (self.x*transf.matrix[2][0] + self.y*transf.matrix[2][1] + self.z*transf.matrix[2][2] + transf.matrix[2][3]))
            
    """
    parsing vector from a string
    input: string e.g. (0,2,4)
    output: 3 coordinate values x, y, z  e.g. 0, 2, 4
    """
    @staticmethod
    def readVector(reader):
        if reader.newToken() == parseReader.MINUS:
            neg = '-'
            reader.newToken()
        else:
            neg = ''
            
        x = coordinate(neg+reader.value())
        
        reader.newToken() # comma
        
        if reader.newToken() == parseReader.MINUS:
            neg = '-'
            reader.newToken()
        else:
            neg = ''
        y = coordinate(neg+reader.value())
        
        reader.newToken() # comma
        
        if reader.newToken() == parseReader.MINUS:
            neg = '-'
            reader.newToken()
        else:
            neg = ''
        z = coordinate(neg+reader.value())
        
        reader.newToken() # bracket

        reader.newToken() # next token
        return x,y,z
            