#-*- coding: utf-8 -*-
"""
@author: akidribm, Dy Raelene Ina Bianchi Louise Mendez
"""

# RHINO SPECIFIC IMPORTS | FROM PYTHON 3.5 to 2.7
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
from __future__ import absolute_import

# Rhino Scripting
import rhinoscriptsyntax as rs
import System
import Rhino as r
import scriptcontext as sc

# Python built-ins
from builtins import range
import os
import warnings

from sortal import coordinate, sort

"""
=============
FUNCTION INFO
=============
'check_precision' returns the precision of comparison (number of decimal places) inside the sortal library (an integer value)

INPUT
None

OUTPUT
Integer value corresponding to number of decimal places the sortal library applies onto numbers

WARNINGS / ERRORS
None

Data Flow: Function call -> Integer value
"""

def check_precision():
    return coordinate.precision()
        
"""
=============
FUNCTION INFO
=============
'set_precision' sets precision of comparison (number of decimal places) inside sortal library (integer value)

INPUT
new: Integer value referring to number of decimal places to indicate precision

OUTPUT
Boolean Value
True, if successful in setting new precision value
False, if unsuccessful

WARNINGS / ERRORS
None

Data Flow: Function call -> Boolean value (True/False)
"""

def set_precision(new):
    if type(new) == type(0):
        coordinate.setPrecision(new) 
        print ('New precision: ', coordinate.precision())
        return True
    else:
        return False

"""
=============
FUNCTION INFO
=============
'default_precision' sets precision of comparison (number of decimal places) inside sortal library back to default (5)   

INPUT
None

OUTPUT
Integer value corresponding to number of decimal places the sortal library applies onto numbers (in this case, 5)

WARNINGS / ERRORS
None

Data Flow: Function call -> Integer value
"""

def default_precision():
    coordinate.setDefaultPrecision()
    return coordinate.precision()
