# -*- coding: utf-8 -*-
"""
Created on Fri Apr 28 14:48:02 2017

@author: akibdpt
"""
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import range
from builtins import super
from builtins import str

from .plane import plane2D, plane3D
from ..struct.coordinate import coordinate, c2, c0, pi
from ..struct.vector import vector2D, vector3D
from ..transform.affineTransform import affineTransform2D, affineTransform3D
from ..io.parseReader import ParsingError


class ellipse2D(plane2D):
    type = 'ind-ellipse2D'
    def __init__(self, sort, focus1, focus2, pin, nonFoci = False, *args, **kwargs):
        super(ellipse2D, self).__init__(sort)     
        
        if focus1 > focus2:
            self.foci = [focus2, focus1]
        else:
            self.foci = [focus1, focus2]
            
        self.center = focus1.midpoint(focus2)
        self.ellipsePin = pin
        dist1 = pin.subtract(self.foci[0])
        dist2 = pin.subtract(self.foci[1])
        
        # radii = [short radius, long radius]
        self.radii = [None,None]
        self.radii[0] = (dist1.dotproduct(dist1).sqrt() + dist2.dotproduct(dist2).sqrt()) / c2
        dist1 = self.foci[1].subtract(self.foci[0])
        self.radii[1] = (self.radii[0] ** c2 - dist1.dotproduct(dist1) / 4).sqrt()
                
    def getPoints(self):
        return self.foci[0],self.foci[1], self.ellipsePin      
        
    def getQuadPoints(self):
        dir1 = self.foci[0].subtract(self.center).unit()
        dir2 = vector2D(dir1.y, c0 - dir1.x)
        return self.center.add(dir1.scale(self.radii[0])), self.center.add(dir2.scale(self.radii[1]))
        
        
    def isNil(self):
        return False
        
    def printVal(self):
        return 'foci:'+str(self.foci[0])+','+str(self.foci[1])+', pin:'+str(self.ellipsePin)
        
    def printRhino(self, dX = 0, dY = 0):
        return '_Ellipse f '+str(self.foci[0].x + dX)+','+str(self.foci[0].y + dY)+' '+str(self.foci[1].x + dX)+','+str(self.foci[1].y + dY)+" "+str(self.ellipsePin.x + dX)+','+str(self.ellipsePin.y + dY)
        
    def toString(self, printKey):
        s = ''
        if printKey and self.key is not None:
            s += str(self.key)+' '
        return s + '<'+str(self.foci[0])+','+str(self.foci[1])+','+str(self.ellipsePin)+'>'
        
    def equals(self, other):
        if self.foci[0] == other.foci[0] and self.foci[1] == other.foci[1]:
            return self.radii[0] == other.radii[0] and self.radii[1] == other.radii[1]
        return False
        
    def containsPoint(self, point):
        dist1 = point.subtract(self.foci[0])
        dist2 = point.subtract(self.foci[1])
        radius1 = ((dist1.dotproduct(dist1)).sqrt() + (dist2.dotproduct(dist2)).sqrt()) / c2
        distFoci = self.foci[1].subtract(self.foci[0])
        if ((radius1 ** c2 - distFoci.dotproduct(distFoci) / 4).sqrt() == self.radii[1] and
            radius1 == self.radii[0]):
                return True
        return False
        
    def area(self):
        return self.radii[0] * self.radii[1] * pi
        
    def compareValue(self, other):
        c = self.foci[0].compareValue(other.foci[0])
        if c != 0:
            return c
        c = self.foci[1].compareValue(other.foci[1])
        if c != 0:
            return c
        c = self.radii[0].compareValue(other.radii[0])
        if c != 0:
            return c
        return self.radii[1].compareValue(other.radii[1])
            
    def concentric(self, other):
        if other.type != 'ind-ellipse2D':
            return False
        return self.foci[0] == other.foci[0] and self.foci[1] == other.foci[1]
        
    def transformInd(self, transformSet):
        affine = transformSet.retrieveTransform(affineTransform2D)
        if affine is None:
            return self.duplicateInd()
        focus1 = self.foci[0].transform(affine)
        focus2 = self.foci[1].transform(affine)
        pin = self.ellipsePin.transform(affine)
        return ellipse2D(self.sort, focus1, focus2, pin)
        
    def replicateInd(self):
        return ellipse2D(self.sort, self.foci[0], self.foci[1], self.ellipsePin)
        
    @staticmethod
    def parse(reader, sort):
        if reader.token() != '<':
            raise ParsingError(reader,'expected a "<" at the beginning of creating a circle')
        if reader.newToken() != '(': #bracket
            raise ParsingError(reader,'expected a "(" ')
        x,y = vector2D.readVector(reader)
        focus1 = vector2D(x,y)
        if reader.token() != ',':
            raise ParsingError(reader,'"," expected')
        reader.newToken() # bracket (current token would be ',')
        x,y = vector2D.readVector(reader)
        focus2 = vector2D(x,y)
        if reader.token() != ',':
            raise ParsingError(reader,'"," expected')
        reader.newToken() # bracket (current token would be ',')
        x,y = vector2D.readVector(reader)
        pin = vector2D(x,y)
        if reader.token() != '>':
            raise ParsingError(reader,'">" expected')
        reader.newToken() #next token
        return ellipse2D(sort, focus1, focus2, pin)
        
class ellipse3D(plane3D):
    type = 'ind-ellipse3D'
    
    def __init__(self, sort, focus1, focus2, pin, *args, **kwargs):
        super(ellipse3D, self).__init__(sort, [focus1, focus2, pin])
        self.ellipsePin = pin
        self.localPin = self.localize(pin)
        
        localFocus1 = self.localize(focus1)
        localFocus2 = self.localize(focus2)
        
        if focus1 > focus2:
            self.localFoci = [localFocus2, localFocus1]
            self.foci = [focus2, focus1]
        else:
            self.localFoci = [localFocus1, localFocus2]
            self.foci = [focus1, focus2]
            
        self.center = focus1.midpoint(focus2)
        self.localCenter = localFocus1.midpoint(localFocus2)
        
        dist1 = self.localPin.subtract(self.localFoci[0])
        dist2 = self.localPin.subtract(self.localFoci[1])
        self.radii = [None,None]
        self.radii[0] = (dist1.dotproduct(dist1).sqrt() + dist2.dotproduct(dist2).sqrt()) / c2
        dist1 = self.localFoci[1].subtract(self.localFoci[0])
        self.radii[1] = (self.radii[0] ** c2 - dist1.dotproduct(dist1) / 4).sqrt()

    def getPoints(self):
        return self.foci[0],self.foci[1], self.ellipsePin     
        
    def getQuadPoints(self):
        dir1 = self.foci[0].subtract(self.center)
        dir2 = dir1.product(self.normal).unit()
        dir1 = dir1.unit()
        return self.center.add(dir1.scale(self.radii[0])), self.center.add(dir2.scale(self.radii[1]))
                
    def isNil(self):
        return False
        
    def printVal(self):
        return ('foci:'+str(self.foci[0])+','+str(self.foci[1])+
                ', pin:'+str(self.ellipsePin))
    
    def printRhino(self, dX = 0, dY = 0):
        f0 = self.foci[0]
        f1 = self.foci[1]
        s  = self.ellipsePin
        return '_Ellipse f '+str(f0.x + dX)+','+str(f0.y + dY)+','+str(f0.z)+' '+str(f1.x + dX)+','+str(f1.y + dY)+','+str(f1.z)+' '+str(s.x + dX)+','+str(s.y + dY)+','+str(s.z)
        
    def toString(self, printKey):
        s = ''
        if printKey and self.key is not None:
            s += str(self.key)+' '
        return s + '<'+str(self.foci[0])+','+str(self.foci[1])+','+str(self.ellipsePin)+'>'
        
    def equals(self, other):
        if self.localFoci[0] == other.localFoci[0] and self.localFoci[1] == other.localFoci[1]:
            return self.radii[0] == other.radii[0] and self.radii[1] == other.radii[1]
        return False
        
    def area(self):
        return self.radii[0] * self.radii[1] * pi
    
    def containsPoint(self, point):
        fc = self.foci[0]
        if point == fc or not point.subtract(fc).dotproduct(self.normal) == 0:
            return False
        p = self.localize(point)
        dist1 = p.subtract(self.localFoci[0])
        dist2 = p.subtract(self.localFoci[1])
        radius1 = (dist1.dotproduct(dist1).sqrt() + dist2.dotproduct(dist2).sqrt()) / c2
        distFoci = self.localFoci[1].subtract(self.localFoci[0])
        if ((radius1 ** c2 - distFoci.dotproduct(distFoci) / 4).sqrt() == self.radii[1] and
            radius1 == self.radii[0]):
                return True
        return False

        """
        dist1 = p.subtract(self.localFoci[0])
        dist2 = p.subtract(self.localFoci[1])
        squareLength = dist1.dotproduct(dist1) + dist2.dotproduct(dist2)
        distFoci = self.localFoci[1].subtract(self.localFoci[0])
        if ((squareLength - distFoci.dotproduct(distFoci)) == self.radii[1] and
            squareLength == self.radii[0]):
                return True
        return False
        """
        
        
        
    
    def compareValue(self, other):
        c = super(ellipse3D, self).compareValue(other)
        if c != 0:
            return c
        c = self.localFoci[0].compareValue(other.localFoci[0])
        if c != 0:
            return c
        c = self.localFoci[1].compareValue(other.localFoci[1])
        if c != 0:
            return c
        c = self.radii[0].compareValue(other.radii[0])
        if c != 0:
            return c
        return self.radii[1].compareValue(other.radii[1])
        
        
    def concentric(self, other):
        if other.type != 'ind-ellipse3D':
            return False
        if not super(ellipse3D, self).__eq__(other):
            return False
        return self.localFoci[0] == other.localFoci[0] and self.localFoci[1] == other.localFoci[1]
        
    def transformInd(self, transformSet):
        affine = transformSet.retrieveTransform(affineTransform3D)
        if affine is None:
            return self.duplicateInd()        
        focus1 = self.foci[0].transform(affine)
        focus2 = self.foci[1].transform(affine)
        pin = self.ellipsePin.transform(affine)
        return ellipse3D(self.sort, focus1, focus2, pin)
        
    def replicateInd(self):
        return ellipse3D(self.sort, self.foci[0], self.foci[1], self.ellipsePin)
        
    @staticmethod
    def parse(reader, sort):
        if reader.token() != '<':
            raise ParsingError(reader,'expected a "<" at the beginning of creating a circle')
        if reader.newToken() != '(': #bracket
            raise ParsingError(reader,'expected a "(" ')
        x,y,z = vector3D.readVector(reader)
        focus1 = vector3D(x,y,z)
        if reader.token() != ',':
            raise ParsingError(reader,'"," expected')
        reader.newToken() # bracket (current token would be ',')
        x,y,z = vector3D.readVector(reader)
        focus2 = vector3D(x,y,z)
        if reader.token() != ',':
            raise ParsingError(reader,'"," expected')
        reader.newToken() # bracket (current token would be ',')
        x,y,z = vector3D.readVector(reader)
        pin = vector3D(x,y,z)
        if reader.token() != '>':
            raise ParsingError(reader,'">" expected')
        reader.newToken() #next token
        return ellipse3D(sort, focus1, focus2, pin)