#-*- coding: utf-8 -*-
"""
@author: akidribm, Dy Raelene Ina Bianchi Louise Mendez
"""

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from sortal import (primitiveSort, point2D, point3D, pointP2D, pointP3D, 
                    lineSegment2D, lineSegment3D, lineSegmentP2D, lineSegmentP3D, 
                    planeSegment2D, planeSegment3D, planeSegmentP2D, planeSegmentP3D,
                    circle2D, circle3D, circleP2D, circleP3D, ellipse2D, ellipse3D, ellipseP2D, ellipseP3D,
                    circularArc2D, circularArc3D, ellipticalArc2D, ellipticalArc3D,
                    bezierCurve2D, bezierCurve3D, bezierCurveP2D, bezierCurveP3D,
                    color, label, rWeight, enumerative, disjunctiveSort, 
                    attributeSort,  description, sort)
                    
"""
=========
FILE INFO
=========

Defines primitive sort types
"""

arg = '{black,white}[black white white white]'
enumSort = primitiveSort('enumerative', enumerative, arg)

# Weight attribute definitions
# rWeight or weight can be used, with the crucial difference being
# rWeight - arithmetic operations on values
# weight - ordinal operations on values
# rWeight does not require an upper limit, but one may be specified
# This upper limit weight value is the value all weightSort individuals are maximised to 
# when any value above the former is stored in the sortal library
weightSort = primitiveSort('weight', rWeight)
weight255Sort = primitiveSort('weight255', rWeight, '255')

# Other attribute types
labelSort = primitiveSort('labels', label)
descSort = primitiveSort('description', description)
ptDSort = primitiveSort('ptD', description)
lsDSort = primitiveSort('lsD', description)
plDSort = primitiveSort('plD', description)
cDSort = primitiveSort('cD', description)
eDSort = primitiveSort('eD', description)
cArcDSort = primitiveSort('cArcD', description)
eArcDSort = primitiveSort('eArcD', description)
bezDSort = primitiveSort('bezD', description)
colorSort = primitiveSort('color', color, 'RGB RGB_SUM,')

# 2D SORTS
# Parametric geometric sorts
pointSortP2D = primitiveSort('point-P2D', pointP2D)
lineSegSortP2D = primitiveSort('lineSeg-P2D', lineSegmentP2D)
planeSegSortP2D = primitiveSort('planeSeg-P2D', planeSegmentP2D) 
cP2D = primitiveSort('circle-P2D', circleP2D)
#cArcP2D = primitiveSort('cArcP2D', circularArcP2D)
eP2D = primitiveSort('ellipse-P2D', ellipseP2D)
#eArcP2D = primitiveSort('eArcP2D', ellipticalArcP2D)
bezP2D = primitiveSort('bezier-P2D', bezierCurveP2D)

# Non-parametric geometric sorts
pointSort2D = primitiveSort('point-N2D', point2D)
lineSegSort2D = primitiveSort('lineSeg-N2D', lineSegment2D)
planeSegSort2D = primitiveSort('planeSeg-N2D', planeSegment2D) 
c2D = primitiveSort('circle-N2D', circle2D)
cArc2D = primitiveSort('cArc-N2D', circularArc2D)
e2D = primitiveSort('ellipse-N2D', ellipse2D)
eArc2D = primitiveSort('eArc-N2D', ellipticalArc2D)
bez2D = primitiveSort('bezier-N2D', bezierCurve2D)

# 3D SORTS
# Parametric geometric sorts
pointSortP3D = primitiveSort('point-P3D', pointP3D)
lineSegSortP3D = primitiveSort('lineSeg-P3D', lineSegmentP3D)
planeSegSortP3D = primitiveSort('planeSeg-P3D', planeSegmentP3D) 
cP3D = primitiveSort('circle-P3D', circleP3D)
#cArcP3D = primitiveSort('cArcP3D', circularArcP3D)
eP3D = primitiveSort('ellipse-P3D', ellipseP3D)
#eArcP3D = primitiveSort('eArcP3D', ellipticalArcP3D)
bezP3D = primitiveSort('bezier-P3D', bezierCurveP3D)

# Non-parametric geometric sorts
pointSort3D = primitiveSort('point-N3D', point3D)
lineSegSort3D = primitiveSort('lineSeg-N3D', lineSegment3D)
planeSegSort3D = primitiveSort('planeSeg-N3D', planeSegment3D) 
c3D = primitiveSort('circle-N3D', circle3D)
cArc3D = primitiveSort('cArc-N3D', circularArc3D)
e3D = primitiveSort('ellipse-N3D', ellipse3D)
eArc3D = primitiveSort('eArc-N3D', ellipticalArc3D)
bez3D = primitiveSort('bezier-N3D', bezierCurve3D)

